"""Extended version of prompt_toolkit's renderer."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.data_structures import Point, Size
from prompt_toolkit.filters import to_filter
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.layout.screen import Char, Screen, WritePosition
from prompt_toolkit.output import ColorDepth, Output
from prompt_toolkit.renderer import Renderer as PtkRenderer
from prompt_toolkit.renderer import _StyleStringHasStyleCache, _StyleStringToAttrsCache

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Optional, Tuple

    from prompt_toolkit.application import Application
    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.layout.layout import Layout
    from prompt_toolkit.styles import BaseStyle


__all__ = ["Renderer"]

log = logging.getLogger(__name__)


def _output_screen_diff(
    app: "Application[Any]",
    output: "Output",
    screen: "Screen",
    current_pos: "Point",
    color_depth: "ColorDepth",
    previous_screen: "Optional[Screen]",
    last_style: "Optional[str]",
    is_done: "bool",  # XXX: drop is_done
    full_screen: "bool",
    attrs_for_style_string: "_StyleStringToAttrsCache",
    style_string_has_style: "_StyleStringHasStyleCache",
    size: "Size",
    previous_width: "int",
) -> "Tuple[Point, Optional[str]]":
    """Render the diff between this screen and the previous screen."""
    width, height = size.columns, size.rows

    #: Variable for capturing the output.
    write = output.write
    write_raw = output.write_raw

    # Create locals for the most used output methods.
    # (Save expensive attribute lookups.)
    _output_set_attributes = output.set_attributes
    _output_reset_attributes = output.reset_attributes
    _output_cursor_forward = output.cursor_forward
    _output_cursor_up = output.cursor_up
    _output_cursor_backward = output.cursor_backward

    # Hide cursor before rendering. (Avoid flickering.)
    output.hide_cursor()

    def reset_attributes() -> None:
        """Wrapper around Output.reset_attributes."""
        nonlocal last_style
        _output_reset_attributes()
        last_style = None  # Forget last char after resetting attributes.

    def move_cursor(new: Point) -> Point:
        """Move cursor to this `new` point & return the given Point."""
        current_x, current_y = current_pos.x, current_pos.y

        if new.y > current_y:
            # Use newlines instead of CURSOR_DOWN, because this might add new lines.
            # CURSOR_DOWN will never create new lines at the bottom.
            # Also reset attributes, otherwise the newline could draw a
            # background color.
            reset_attributes()
            write("\r\n" * (new.y - current_y))
            current_x = 0
            _output_cursor_forward(new.x)
            return new
        elif new.y < current_y:
            _output_cursor_up(current_y - new.y)

        if current_x >= width - 1:
            write("\r")
            _output_cursor_forward(new.x)
        elif new.x < current_x or current_x >= width - 1:
            _output_cursor_backward(current_x - new.x)
        elif new.x > current_x:
            _output_cursor_forward(new.x - current_x)

        return new

    def output_char(char: Char) -> None:
        """Write the output of this character."""
        nonlocal last_style

        # If the last printed character has the same style, don't output the
        # style again.
        if last_style == char.style:
            write(char.char)
        else:
            # Look up `Attr` for this style string. Only set attributes if different.
            # (Two style strings can still have the same formatting.)
            # Note that an empty style string can have formatting that needs to
            # be applied, because of style transformations.
            new_attrs = attrs_for_style_string[char.style]
            if not last_style or new_attrs != attrs_for_style_string[last_style]:
                _output_set_attributes(new_attrs, color_depth)

            write(char.char)
            last_style = char.style

    def get_max_column_index(row: Dict[int, Char], zwe_row: Dict[int, str]) -> int:
        """Return max used column index, ignoring trailing unstyled whitespace."""
        return max(
            {
                index
                for index, cell in row.items()
                if cell.char != " " or style_string_has_style[cell.style]
            }
            | zwe_row.keys()
            | {0}
        )

    # Render for the first time: reset styling.
    if not previous_screen:
        reset_attributes()

    # Disable autowrap. (When entering a the alternate screen, or anytime when
    # we have a prompt. - In the case of a REPL, like IPython, people can have
    # background threads, and it's hard for debugging if their output is not
    # wrapped.)
    if not previous_screen or not full_screen:
        output.disable_autowrap()

    # When the previous screen has a different size, redraw everything anyway.
    # Also when we are done. (We might take up less rows, so clearing is important.)
    if (
        is_done or not previous_screen or previous_width != width
    ):  # XXX: also consider height??
        current_pos = move_cursor(Point(x=0, y=0))
        reset_attributes()
        output.erase_down()

        previous_screen = Screen()

    # Get height of the screen.
    # (height changes as we loop over data_buffer, so remember the current value.)
    # (Also make sure to clip the height to the size of the output.)
    current_height = min(screen.height, height)

    # Loop over the rows.
    row_count = min(max(screen.height, previous_screen.height), height)
    c = 0  # Column counter.

    for y in range(row_count):
        new_row = screen.data_buffer[y]
        previous_row = previous_screen.data_buffer[y]
        zwe_row = screen.zero_width_escapes[y]
        previous_zwe_row = previous_screen.zero_width_escapes[y]

        new_max_line_len = min(width - 1, get_max_column_index(new_row, zwe_row))
        previous_max_line_len = min(
            width - 1, get_max_column_index(previous_row, previous_zwe_row)
        )

        # Loop over the columns.
        c = 0

        prev_diff_char = False

        # We might have a ZWE sequence just beyond the end of the line
        while c <= new_max_line_len:
            new_char = new_row[c]
            old_char = previous_row[c]
            new_zwe = zwe_row[c]
            char_width = new_char.width or 1

            # When the old and new character at this position are different,
            # draw the output. (Because of the performance, we don't call
            # `Char.__ne__`, but inline the same expression.)
            diff_char = (
                new_char.char != old_char.char or new_char.style != old_char.style
            )

            # Redraw escape sequences if the escape sequence at this position changed,
            # or if the current or previous character changed
            if new_zwe != previous_zwe_row[c] or diff_char or prev_diff_char:
                # Send injected escape sequences to output.
                write_raw(new_zwe)

            if diff_char:
                # Don't move the cursor if it is already in the correct position
                if c != current_pos.x or y != current_pos.y:
                    current_pos = move_cursor(Point(x=c, y=y))

                output_char(new_char)
                current_pos = Point(x=current_pos.x + char_width, y=current_pos.y)

            prev_diff_char = diff_char

            c += char_width

        # If the new line is shorter, trim it.
        if previous_screen and new_max_line_len < previous_max_line_len:
            # Don't move the cursor if it is already in the correct position
            if current_pos.x != new_max_line_len or current_pos.y != current_pos.y:
                current_pos = move_cursor(Point(x=new_max_line_len + 1, y=y))
            reset_attributes()
            output.erase_end_of_line()

    # Correctly reserve vertical space as required by the layout.
    # When this is a new screen (drawn for the first time), or for some reason
    # higher than the previous one. Move the cursor once to the bottom of the
    # output. That way, we're sure that the terminal scrolls up, even when the
    # lower lines of the canvas just contain whitespace.

    # The most obvious reason that we actually want this behaviour is the avoid
    # the artifact of the input scrolling when the completion menu is shown.
    # (If the scrolling is actually wanted, the layout can still be build in a
    # way to behave that way by setting a dynamic height.)
    if current_height > previous_screen.height:
        current_pos = move_cursor(Point(x=0, y=current_height - 1))

    # Move cursor:
    if is_done:
        current_pos = move_cursor(Point(x=0, y=current_height))
        output.erase_down()
    else:
        current_pos = move_cursor(screen.get_cursor_position(app.layout.current_window))

    if is_done or not full_screen:
        output.enable_autowrap()

    # Always reset the color attributes. This is important because a background
    # thread could print data to stdout and we want that to be displayed in the
    # default colors. (Also, if a background color has been set, many terminals
    # give weird artifacts on resize events.)
    reset_attributes()

    if screen.show_cursor or is_done:
        output.show_cursor()

    return current_pos, last_style


class Renderer(PtkRenderer):
    """Renderer with modifications."""

    def __init__(
        self,
        style: "BaseStyle",
        output: "Output",
        full_screen: "bool" = False,
        mouse_support: "FilterOrBool" = False,
        cpr_not_supported_callback: "Optional[Callable[[], None]]" = None,
        extend_height: "FilterOrBool" = False,
        extend_width: "FilterOrBool" = False,
    ) -> None:
        """Create a new :py:class:`Renderer` instance."""
        super().__init__(
            style, output, full_screen, mouse_support, cpr_not_supported_callback
        )
        self.extend_height = to_filter(extend_height)
        self.extend_width = to_filter(extend_width)

    def render(
        self, app: "Application[Any]", layout: "Layout", is_done: bool = False
    ) -> None:
        """Render the current interface to the output."""
        output = self.output

        # Enter alternate screen.
        if self.full_screen and not self._in_alternate_screen:
            self._in_alternate_screen = True
            output.enter_alternate_screen()

        # Enable bracketed paste.
        if not self._bracketed_paste_enabled:
            self.output.enable_bracketed_paste()
            self._bracketed_paste_enabled = True

        # Reset cursor key mode.
        if not self._cursor_key_mode_reset:
            self.output.reset_cursor_key_mode()
            self._cursor_key_mode_reset = True

        # Enable/disable mouse support.
        needs_mouse_support = self.mouse_support()

        if needs_mouse_support and not self._mouse_support_enabled:
            output.enable_mouse_support()
            self._mouse_support_enabled = True

        elif not needs_mouse_support and self._mouse_support_enabled:
            output.disable_mouse_support()
            self._mouse_support_enabled = False

        # Create screen and write layout to it.
        size = output.get_size()
        screen = Screen()
        screen.show_cursor = False  # Hide cursor by default, unless one of the
        # containers decides to display it.
        mouse_handlers = MouseHandlers()

        # Calculate height.
        if self.full_screen:
            height = size.rows
        elif is_done:
            # When we are done, we don't necessary want to fill up until the bottom.
            height = layout.container.preferred_height(
                size.columns, size.rows
            ).preferred
        else:
            last_height = self._last_screen.height if self._last_screen else 0
            height = max(
                self._min_available_height,
                last_height,
                layout.container.preferred_height(size.columns, size.rows).preferred,
            )

        height = min(height, size.rows)

        # When the size changes, don't consider the previous screen.
        if self._last_size != size:
            self._last_screen = None

        # When we render using another style or another color depth, do a full
        # repaint. (Forget about the previous rendered screen.)
        # (But note that we still use _last_screen to calculate the height.)
        if (
            self.style.invalidation_hash() != self._last_style_hash
            or app.style_transformation.invalidation_hash()
            != self._last_transformation_hash
            or app.color_depth != self._last_color_depth
        ):
            self._last_screen = None
            self._attrs_for_style = None
            self._style_string_has_style = None

        if self._attrs_for_style is None:
            self._attrs_for_style = _StyleStringToAttrsCache(
                self.style.get_attrs_for_style_str, app.style_transformation
            )
        if self._style_string_has_style is None:
            self._style_string_has_style = _StyleStringHasStyleCache(
                self._attrs_for_style
            )

        self._last_style_hash = self.style.invalidation_hash()
        self._last_transformation_hash = app.style_transformation.invalidation_hash()
        self._last_color_depth = app.color_depth

        layout.container.write_to_screen(
            screen,
            mouse_handlers,
            WritePosition(xpos=0, ypos=0, width=size.columns, height=height),
            parent_style="",
            erase_bg=False,
            z_index=None,
        )
        screen.draw_all_floats()

        # When grayed. Replace all styles in the new screen.
        if app.exit_style:
            screen.append_style_to_content(app.exit_style)

        # Expand size if required
        output_size = size
        if self.extend_height():
            output_size = Size(9999999, output_size.columns)
        if self.extend_width():
            output_size = Size(size.rows, output_size.columns + 1)

        # Process diff and write to output.
        self._cursor_pos, self._last_style = _output_screen_diff(
            app,
            output,
            screen,
            self._cursor_pos,
            app.color_depth,
            self._last_screen,
            self._last_style,
            is_done,
            full_screen=self.full_screen,
            attrs_for_style_string=self._attrs_for_style,
            style_string_has_style=self._style_string_has_style,
            size=output_size,
            previous_width=(self._last_size.columns if self._last_size else 0),
        )
        self._last_screen = screen
        self._last_size = size
        self.mouse_handlers = mouse_handlers

        # Handle cursor shapes.
        new_cursor_shape = app.cursor.get_cursor_shape(app)
        if (
            self._last_cursor_shape is None
            or self._last_cursor_shape != new_cursor_shape
        ):
            output.set_cursor_shape(new_cursor_shape)
            self._last_cursor_shape = new_cursor_shape

        # Flush buffered output.
        output.flush()

        # Set visible windows in layout.
        app.layout.visible_windows = screen.visible_windows

        if is_done:
            self.reset()



"""Utilities for manipulating formatted text."""

from __future__ import annotations

import re
from enum import Enum
from typing import Iterable, Optional, cast

from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple, StyleAndTextTuples
from prompt_toolkit.formatted_text.utils import (
    fragment_list_to_text,
    fragment_list_width,
    split_lines,
)
from prompt_toolkit.utils import get_cwidth
from pygments.lexers import get_lexer_by_name
from pygments.util import ClassNotFound

from euporie.core.border import GridStyle, Padding, Thin


class FormattedTextAlign(Enum):
    """Alignment of formatted text."""

    LEFT = "LEFT"
    RIGHT = "RIGHT"
    CENTER = "CENTER"


def last_line_length(ft: StyleAndTextTuples) -> int:
    """Calculate the length of the last line in formatted text."""
    line: StyleAndTextTuples = []
    for style, text, *_ in ft[::-1]:
        index = text.find("\n")
        line.append((style, text[index + 1 :]))
        if index > -1:
            break
    return fragment_list_width(line)


def max_line_width(ft: StyleAndTextTuples) -> int:
    """Calculate the length of the longest line in formatted text."""
    return max(fragment_list_width(line) for line in split_lines(ft))


def fragment_list_to_words(
    fragments: "StyleAndTextTuples", sep: "str" = " "
) -> "Iterable[OneStyleAndTextTuple]":
    """Split formatted text into word fragments."""
    for style, string, *mouse_handler in fragments:
        # for part in re.split(r"\b", string):
        for part in re.split(r"(?=\s)", string):
            yield cast("OneStyleAndTextTuple", (style, part, *mouse_handler))


def apply_style(ft: StyleAndTextTuples, style: str) -> StyleAndTextTuples:
    """Apply a style to formatted text."""
    return [
        (
            f"{fragment_style} {style}"
            if "[ZeroWidthEscape]" not in fragment_style
            else fragment_style,
            text,
        )
        for (fragment_style, text, *_) in ft
    ]


def strip(
    ft: StyleAndTextTuples,
    left: bool = True,
    right: bool = True,
    char: Optional[str] = None,
    only_unstyled: bool = False,
) -> StyleAndTextTuples:
    """Strip whitespace (or a given character) from the ends of formatted text.

    Args:
        ft: The formatted text to strip
        left: If :py:const:`True`, strip from the left side of the input
        right: If :py:const:`True`, strip from the right side of the input
        char: The character to strip. If :py:const:`None`, strips whitespace
        only_unstyled: If :py:const:`True`, only strip unstyled fragments

    Returns:
        The stripped formatted text

    """
    result = ft[:]
    for toggle, index, strip_func in [(left, 0, str.lstrip), (right, -1, str.rstrip)]:
        if result and toggle:
            text = strip_func(result[index][1], char)
            while result and not text:
                del result[index]
                if not result:
                    break
                text = strip_func(result[index][1], char)
            if result and "[ZeroWidthEscape]" not in result[index][0]:
                result[index] = (result[index][0], text)
    return result


def truncate(
    ft: StyleAndTextTuples,
    width: int,
    style: str = "",
    placeholder: str = "…",
) -> StyleAndTextTuples:
    """Truncates all lines at a given length.

    Args:
        ft: The formatted text to truncate
        width: The width at which to truncate the text
        style: The style to apply to the truncation placeholder. The style of the
            truncated text will be used if not provided
        placeholder: The string that will appear at the end of a truncated line

    Returns:
        The truncated formatted text

    """
    result: StyleAndTextTuples = []
    phw = sum(get_cwidth(c) for c in placeholder)
    for line in split_lines(ft):
        used_width = 0
        for item in line:
            fragment_width = sum(
                get_cwidth(c) for c in item[1] if "[ZeroWidthEscape]" not in item[0]
            )
            if used_width + fragment_width > width - phw:
                remaining_width = width - used_width - fragment_width - phw
                result.append((item[0], item[1][:remaining_width]))
                result.append((style or item[0], placeholder))
                break
            else:
                result.append(item)
                used_width += fragment_width
        result.append(("", "\n"))
    result.pop()
    return result


def wrap(
    ft: StyleAndTextTuples,
    width: int,
    style: str = "",
    placeholder: str = "…",
    left: "int" = 0,
    truncate_long_words: "bool" = True,
    strip_trailing_ws: "bool" = False,
) -> StyleAndTextTuples:
    """Wraps formatted text at a given width.

    If words are longer than the given line they will be truncated

    Args:
        ft: The formatted text to wrap
        width: The width at which to wrap the text
        style: The style to apply to the truncation placeholder
        placeholder: The string that will appear at the end of a truncated line
        left: The starting position within the first list
        truncate_long_words: If :const:`True` words longer than a line will be
            truncated
        strip_trailing_ws: If :const:`True`, trailing whitespace will be removed from
            the ends of lines

    Returns:
        The wrapped formatted text
    """
    result: StyleAndTextTuples = []
    lines = list(split_lines(ft))
    output_line = 0
    for i, line in enumerate(lines):
        if fragment_list_width(line) <= width - left:
            result += line
            if i < len(lines) - 1:
                result.append(("", "\n"))
            output_line += 1
            left = 0
        else:
            for item in fragment_list_to_words(line):
                # Skip empty fragments
                if item[1] == "":
                    continue
                fragment_width = sum(
                    get_cwidth(c) for c in item[1] if "[ZeroWidthEscape]" not in item[0]
                )
                # Start a new line - we are at the end of the current output line
                if left + fragment_width > width and left > 0:

                    # Add as much trailing whitespace as we can
                    if not strip_trailing_ws and (trailing_ws := (not item[1].strip())):
                        result.append((item[0], item[1][: width - left]))

                    # Remove trailing whitespace
                    if strip_trailing_ws:
                        result = strip(result, left=False)

                    # Start new line
                    result.append(("", "\n"))
                    output_line += 1
                    left = 0

                    # If we added trailing whitespace, process the next fragment
                    if not strip_trailing_ws and trailing_ws:
                        continue

                # Strip left-hand whitespace from a word at the start of a line
                # if output_line != 0 and left == 0:
                if left == 0:
                    item = (item[0], item[1].lstrip(" "))
                # Truncate words longer than a line
                if truncate and left == 0 and fragment_width > width - left:
                    result += truncate([item], width - left, style, placeholder)
                    left += fragment_width
                # Otherwise just add the word to the line
                else:
                    result.append(item)
                    left += fragment_width

    if strip_trailing_ws:
        result = strip(result, left=False)

    return result


def align(
    how: FormattedTextAlign,
    ft: StyleAndTextTuples,
    width: Optional[int] = None,
    style: str = "",
    placeholder: str = "…",
) -> StyleAndTextTuples:
    """Align formatted text at a given width.

    Args:
        how: The alignment direction
        ft: The formatted text to strip
        width: The width to which the output should be padded. If :py:const:`None`, the
            length of the longest line is used
        style: The style to apply to the padding
        placeholder: The string that will appear at the end of a truncated line

    Returns:
        The aligned formatted text

    """
    style = f"{style} nounderline"
    lines = split_lines(ft)
    if width is None:
        lines = [strip(line) for line in split_lines(ft)]
        width = max(fragment_list_width(line) for line in lines)
    result: StyleAndTextTuples = []
    for line in lines:
        line_width = fragment_list_width(line)
        # Truncate the line if it is too long
        if line_width > width:
            result += truncate(line, width, style, placeholder)
        else:
            pad_left = pad_right = 0
            if how == FormattedTextAlign.CENTER:
                pad_left = (width - line_width) // 2
                pad_right = width - line_width - pad_left
            elif how == FormattedTextAlign.LEFT:
                pad_right = width - line_width
            elif how == FormattedTextAlign.RIGHT:
                pad_left = width - line_width
            if pad_left:
                result.append((style, " " * pad_left))
            result += line
            if pad_right:
                result.append((style, " " * pad_right))
        result.append((style, "\n"))
    result.pop()
    return result


def indent(
    ft: StyleAndTextTuples,
    margin: str = " ",
    style: str = "",
    skip_first: bool = False,
) -> StyleAndTextTuples:
    """Indents formatted text with a given margin.

    Args:
        ft: The formatted text to strip
        margin: The margin string to add
        style: The style to apply to the margin
        skip_first: If :py:const:`True`, the first line is skipped

    Returns:
        The indented formatted text

    """
    result: StyleAndTextTuples = []
    for i, line in enumerate(split_lines(ft)):
        if not (i == 0 and skip_first):
            result.append((style, margin))
        result += line
        result.append(("", "\n"))
    result.pop()
    return result


def add_border(
    ft: StyleAndTextTuples,
    width: "Optional[int]" = None,
    style: str = "",
    border: "GridStyle" = Thin.grid,
    padding: "Optional[Padding]" = None,
) -> StyleAndTextTuples:
    """Adds a border around formatted text.

    Args:
        ft: The formatted text to enclose with a border
        width: The target width including the border and padding
        style: The style to apply to the border
        border: The grid style to use for the border
        padding: The width of spacing to apply between the content and the border

    Returns:
        The indented formatted text

    """
    if padding is None:
        padding = Padding(0, 1, 0, 1)
    if isinstance(padding, int):
        padding = Padding(padding, padding, padding, padding)
    if len(padding) == 2:
        padding = Padding(padding[0], padding[1], padding[0], padding[1])
    # `None` is not permitted for padding here
    padding = Padding(
        padding[0] or 0, padding[1] or 0, padding[2] or 0, padding[3] or 0
    )

    max_lw = max_line_width(ft)
    if width is None:
        width = max_lw + (padding.right or 0) + (padding.left or 0) + 2

    inner_width = width - 2
    # Ensure all lines are the same length
    ft = align(
        FormattedTextAlign.LEFT,
        ft,
        width=inner_width - (padding.right or 0) - (padding.left or 0),
        style=style,
    )

    result: StyleAndTextTuples = []

    style = f"{style} nounderline"

    result.extend(
        [
            (
                f"{style} class:border,left,top",
                border.TOP_LEFT,
            ),
            (
                f"{style} class:border,top",
                border.TOP_MID * inner_width,
            ),
            (
                f"{style} class:border,right,top",
                border.TOP_RIGHT + "\n",
            ),
        ]
    )
    for _ in range(padding.top or 0):
        result.extend(
            [
                (f"{style} class:border,left", border.MID_LEFT),
                (style, " " * inner_width),
                (f"{style} class:border,right", border.MID_RIGHT + "\n"),
            ]
        )
    for line in split_lines(ft):
        result.extend(
            [
                (f"{style} class:border,left", border.MID_LEFT),
                (style, " " * (padding.left or 0)),
                *line,
                (style, " " * (padding.right or 0)),
                (f"{style} class:border,right", border.MID_RIGHT + "\n"),
            ]
        )
    for _ in range(padding.bottom or 0):
        result.extend(
            [
                (f"{style} class:border,left", border.MID_LEFT),
                (style, " " * inner_width),
                (f"{style} class:border,right", border.MID_RIGHT + "\n"),
            ]
        )
    result.extend(
        [
            (
                f"{style} class:border,left,bottom",
                border.BOTTOM_LEFT,
            ),
            (
                f"{style} class:border,bottom",
                border.BOTTOM_MID * inner_width,
            ),
            (
                f"{style} class:border,right,bottom",
                border.BOTTOM_RIGHT + "\n",
            ),
        ]
    )
    return result


def lex(ft: "StyleAndTextTuples", lexer_name: "str") -> "StyleAndTextTuples":
    """Format formatted text using a named :py:mod:`pygments` lexer."""
    from prompt_toolkit.lexers.pygments import _token_cache

    text = fragment_list_to_text(ft)
    try:
        lexer = get_lexer_by_name(lexer_name)
    except ClassNotFound:
        return ft
    else:
        return [(_token_cache[t], v) for _, t, v in lexer.get_tokens_unprocessed(text)]


def paste(
    ft_bottom: "StyleAndTextTuples",
    ft_top: "StyleAndTextTuples",
    row: "int" = 0,
    col: "int" = 0,
) -> "StyleAndTextTuples":
    """Paste formatted text on top of other formatted text."""
    ft: "StyleAndTextTuples" = []

    top_lines = dict(enumerate(split_lines(ft_top), start=row))
    for y, line_b in enumerate(split_lines(ft_bottom)):
        if y in top_lines:
            x = 0
            line_t = top_lines[y]
            line_t_width = fragment_list_width(line_t)
            done = False
            for frag in line_b:
                fragment_width = sum(
                    get_cwidth(c) for c in frag[1] if "[ZeroWidthEscape]" not in frag[0]
                )
                if x < col:
                    remaining = col - x
                    ft.append((frag[0], frag[1][:remaining], *frag[2:]))
                elif x == col and not done:
                    ft += line_t
                    done = True
                elif x >= col + line_t_width:
                    ft.append(frag)
                x += fragment_width
        else:
            ft += line_b
        ft.append(("", "\n"))

    if ft:
        ft.pop()

    return ft



"""Contains functions to automatically format code cell input."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from euporie.core.log import stdout_to_log

if TYPE_CHECKING:
    from euporie.core.config import Config

log = logging.getLogger(__name__)


def format_black(text: "str") -> "str":
    """Format a code string using :py:mod:`black`."""
    try:
        import black
    except ModuleNotFoundError:
        pass
    else:
        try:
            text = black.format_str(text, mode=black.Mode()).rstrip()
        except black.parsing.InvalidInput:
            log.warning("Error formatting code with black: invalid input")
    return text


def format_isort(text: "str") -> "str":
    """Format a code string using :py:mod:`isort`."""
    try:
        import isort
    except ModuleNotFoundError:
        pass
    else:
        text = isort.code(text, profile="black")
    return text


def format_ssort(text: "str") -> "str":
    """Format a code string using :py:mod:`ssort`."""
    try:
        import ssort
    except ModuleNotFoundError:
        pass
    else:
        with stdout_to_log(log, output="stderr"):
            output = ""
            try:
                output = ssort.ssort(text)
            except Exception:  # noqa S110
                # log.debug("Error formatting code with ssort")
                pass
            finally:
                if output:
                    text = output
    return text


def format_code(text: "str", config: "Config") -> "str":
    """Format a code string using :py:mod:``."""
    if config.format_ssort:
        text = format_ssort(text)
    if config.format_isort:
        text = format_isort(text)
    if config.format_black:
        text = format_black(text)
    return text.strip()



"""Contains margins."""

from __future__ import annotations

import asyncio
import logging
from abc import ABCMeta
from typing import TYPE_CHECKING, cast

from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import Condition, FilterOrBool, to_filter
from prompt_toolkit.layout.containers import Container, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.margins import Margin
from prompt_toolkit.mouse_events import MouseButton
from prompt_toolkit.mouse_events import MouseEvent as PtkMouseEvent
from prompt_toolkit.mouse_events import MouseEventType

from euporie.core.current import get_app
from euporie.core.key_binding.bindings.mouse import MouseEvent, RelativePosition

if TYPE_CHECKING:
    from typing import Callable, Optional, Protocol

    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.layout.containers import WindowRenderInfo
    from prompt_toolkit.layout.controls import UIContent
    from prompt_toolkit.layout.mouse_handlers import MouseHandlers
    from prompt_toolkit.layout.screen import Screen, WritePosition

    class ScrollableContainer(Protocol):
        """Protocol for a scrollable container."""

        render_info: WindowRenderInfo
        vertical_scroll: int


log = logging.getLogger(__name__)


class ClickableMargin(Margin, metaclass=ABCMeta):
    """A margin sub-class which handles mouse events."""

    write_position: "Optional[WritePosition]"

    def set_write_position(self, write_position: "WritePosition") -> "None":
        """Set the write position of the menu."""
        self.write_position = write_position


class MarginContainer(Window):
    """A container which renders a stand-alone margin."""

    def __init__(self, margin: "Margin", target: "ScrollableContainer") -> "None":
        """Create a new instance."""
        self.margin = margin
        self.target = target
        self.render_info: "Optional[WindowRenderInfo]" = None
        self.content = FormattedTextControl(self.create_fragments)

    def create_fragments(self) -> "StyleAndTextTuples":
        """Generate text fragments to display."""
        return self.margin.create_margin(
            self.target.render_info,
            self.write_position.width,
            self.write_position.height,
        )

    def reset(self) -> "None":
        """Reset the state of this container and all the children."""

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Return a the desired width for this container."""
        width = self.margin.get_width(lambda: self.target.render_info.ui_content)
        return Dimension(min=width, max=width)

    def preferred_height(
        self, width: "int", max_available_height: "int"
    ) -> "Dimension":
        """Return a thedesired height for this container."""
        return Dimension(max_available_height)

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: str,
        erase_bg: bool,
        z_index: Optional[int],
    ) -> None:
        """Write the actual content to the screen."""
        self.write_position = write_position
        if isinstance(self.margin, ClickableMargin):
            self.margin.set_write_position(write_position)

        margin_content: "UIContent" = self.content.create_content(
            write_position.width + 1, write_position.height
        )
        visible_line_to_row_col, rowcol_to_yx = self._copy_body(
            margin_content, screen, write_position, 0, write_position.width
        )

        # Set mouse handlers.
        def mouse_handler(mouse_event: PtkMouseEvent) -> "NotImplementedOrNone":
            """Turns screen coordinates into line coordinates."""
            # Don't handle mouse events outside of the current modal part of the UI
            if self not in get_app().layout.walk_through_modal_area():
                return NotImplemented

            # Find row/col position first.
            yx_to_rowcol = {v: k for k, v in rowcol_to_yx.items()}
            y = mouse_event.position.y
            x = mouse_event.position.x

            # If clicked below the content area, look for a position in the
            # last line instead
            max_y = write_position.ypos + len(visible_line_to_row_col) - 1
            y = min(max_y, y)
            result: "NotImplementedOrNone" = NotImplemented

            while x >= 0:
                try:
                    row, col = yx_to_rowcol[y, x]
                except KeyError:
                    # Try again. (When clicking on the right side of double
                    # width characters, or on the right side of the input.)
                    x -= 1
                else:
                    # Found position, call handler of UIControl.
                    if isinstance(mouse_event, MouseEvent):
                        cell_position = mouse_event.cell_position
                    else:
                        cell_position = RelativePosition(0.5, 0.5)
                    result = self.content.mouse_handler(
                        MouseEvent(
                            position=Point(x=col, y=row),
                            event_type=mouse_event.event_type,
                            button=mouse_event.button,
                            modifiers=mouse_event.modifiers,
                            cell_position=cell_position,
                        )
                    )
                    break

            return result

        mouse_handlers.set_mouse_handler_for_range(
            x_min=write_position.xpos,
            x_max=write_position.xpos + write_position.width,
            y_min=write_position.ypos,
            y_max=write_position.ypos + write_position.height,
            handler=mouse_handler,
        )

    def is_modal(self) -> "bool":
        """When this container is modal."""
        return False

    def get_key_bindings(self) -> "Optional[KeyBindingsBase]":
        """Returns a :class:`.KeyBindings` object."""
        return None

    def get_children(self) -> "list[Container]":
        """Return the list of child :class:`.Container` objects."""
        return []


class ScrollbarMargin(ClickableMargin):
    """Margin displaying a scrollbar.

    Args:
        display_arrows: Display scroll up/down arrows.
        up_arrow: Character to use for the scrollbar's up arrow
        down_arrow: Character to use for the scrollbar's down arrow
        smooth: Use block character to move scrollbar more smoothly

    """

    eighths = "█▇▆▅▄▃▂▁ "

    def __init__(
        self,
        display_arrows: "FilterOrBool" = True,
        up_arrow_symbol: "str" = "▴",
        down_arrow_symbol: "str" = "▾",
        autohide: "FilterOrBool" = True,
        smooth: "bool" = True,
        style: "str" = "",
    ) -> "None":
        """Creates a new scrollbar instance."""
        self.autohide = to_filter(autohide)
        self.display_arrows = to_filter(display_arrows)
        self.up_arrow_symbol = up_arrow_symbol
        self.down_arrow_symbol = down_arrow_symbol
        self.smooth = smooth
        self.style = style

        self.repeat_task: "Optional[asyncio.Task[None]]" = None
        self.dragging = False
        self.drag_start_scroll = 0
        self.drag_start_offset = 0.0

        self.thumb_top = 0.0
        self.thumb_size = 0.0

        self.window_render_info: "Optional[WindowRenderInfo]" = None
        self.write_position: "Optional[WritePosition]" = None

        self.visible = (
            self.autohide
            & Condition(
                lambda: self.window_render_info is None
                or (
                    self.window_render_info is not None
                    and self.window_render_info.content_height
                    > self.window_render_info.window_height
                )
            )
        ) | ~self.autohide

    def get_width(self, get_ui_content: "Callable[[], UIContent]") -> "int":
        """Return the scrollbar width: always 1."""
        return 1 if self.visible() else 0

    def create_margin(
        self,
        window_render_info: "WindowRenderInfo",
        width: "int",
        height: "int",
        margin_render_info: "Optional[WindowRenderInfo]" = None,
    ) -> "StyleAndTextTuples":
        """Creates the margin's formatted text."""
        result: StyleAndTextTuples = []

        self.window_render_info = window_render_info
        self.margin_render_info = margin_render_info

        if not width or not self.visible():
            return result

        # Show we render the arrow buttons?
        display_arrows = self.display_arrows()

        # The height of the scrollbar, excluding the optional buttons
        self.track_height = window_render_info.window_height
        if display_arrows:
            self.track_height -= 2

        # Height of all text in the output: If there is none, we cannot divide
        # by zero so we hide the thumb
        content_height = window_render_info.content_height
        if content_height == 0 or content_height <= len(
            window_render_info.displayed_lines
        ):
            self.thumb_size = 0
        else:
            # The thumb is the part which moves, floating on the track: calculate its size
            fraction_visible = len(window_render_info.displayed_lines) / (
                content_height
            )
            self.thumb_size = (
                int(
                    min(self.track_height, max(1, self.track_height * fraction_visible))
                    * 8
                )
                / 8
            )
        if not self.smooth:
            self.thumb_size = int(self.thumb_size)

        # Calculate the position of the thumb
        if content_height <= len(window_render_info.displayed_lines):
            fraction_above = 0.0
        else:
            fraction_above = window_render_info.vertical_scroll / (
                content_height - len(window_render_info.displayed_lines)
            )
        self.thumb_top = max(
            0,
            min(
                self.track_height - self.thumb_size,
                (int((self.track_height - self.thumb_size) * fraction_above * 8) / 8),
            ),
        )
        if not self.smooth:
            self.thumb_top = int(self.thumb_top)

        # Determine which characters to use for the ends of the thumb
        thumb_top_char = self.eighths[int(self.thumb_top % 1 * 8)]
        thumb_bottom_char = self.eighths[
            int((self.thumb_top + self.thumb_size) % 1 * 8)
        ]

        # Calculate thumb dimensions
        show_thumb_top = (self.thumb_top % 1) != 0
        thumb_top_size = 1 - self.thumb_top % 1
        show_thumb_bottom = (self.thumb_top + self.thumb_size) % 1 != 0
        thumb_bottom_size = (self.thumb_top + self.thumb_size) % 1
        thumb_middle_size = int(
            self.thumb_size
            - show_thumb_top * thumb_top_size
            - show_thumb_bottom * thumb_bottom_size
        )
        rows_after_thumb = (
            self.track_height
            - int(self.thumb_top)
            - show_thumb_top
            - thumb_middle_size
            - show_thumb_bottom
        )

        # Construct the scrollbar

        mouse_handler = cast("Callable[[PtkMouseEvent], None]", self.mouse_handler)

        # Up button
        if display_arrows:
            result += [
                ("class:scrollbar.arrow", self.up_arrow_symbol, mouse_handler),
                ("class:scrollbar", "\n", mouse_handler),
            ]
        # Track above the thumb
        for _ in range(int(self.thumb_top)):
            result += [
                ("class:scrollbar.background", " ", mouse_handler),
                ("class:scrollbar", "\n", mouse_handler),
            ]
        # Top of thumb
        if show_thumb_top:
            result += [
                (
                    "class:scrollbar.background,scrollbar.start",
                    thumb_top_char,
                    mouse_handler,
                ),
                ("class:scrollbar", "\n", mouse_handler),
            ]
        # Middle of thumb
        for _ in range(thumb_middle_size):
            result += [
                ("class:scrollbar.button", " ", mouse_handler),
                ("class:scrollbar", "\n", mouse_handler),
            ]
        # Bottom of thumb
        if show_thumb_bottom:
            result += [
                (
                    "class:scrollbar.background,scrollbar.end",
                    thumb_bottom_char,
                    mouse_handler,
                ),
                ("class:scrollbar", "\n", mouse_handler),
            ]
        # Track below the thumb
        for _ in range(rows_after_thumb):
            result += [
                ("class:scrollbar.background", " ", mouse_handler),
                ("class:scrollbar", "\n", mouse_handler),
            ]
        # Down button
        if display_arrows:
            result += [
                ("class:scrollbar.arrow", self.down_arrow_symbol, mouse_handler),
            ]

        # if self.style:
        # result = [
        # (f"{self.style} {style}", text, *cb) for style, text, *cb in result
        # ]

        return result

    def _mouse_handler(
        self, mouse_event: "PtkMouseEvent", repeated: "bool" = False
    ) -> "NotImplementedOrNone":
        """Handle scrollbar mouse events.

        Scrolls up or down if the arrows are clicked, repeating while the mouse button
        is held down. Scolls up or down one page if the background is clicked,
        repeating while the left mouse button is held down. Scrolls if the
        scrollbar-button is dragged. Scrolls if the scroll-wheel is used on the
        scrollbar.

        Args:
            mouse_event: The triggering mouse event
            repeated: Set to True if the method is running as a repeated event

        Returns:
            :py:const:`NotImplemented` is eturned when the mouse event is unhandled;
            :py:const:`None` is returned when the mouse event is handled successfully

        """
        render_info = self.window_render_info
        assert render_info is not None

        content_height = render_info.content_height
        if isinstance(mouse_event, MouseEvent):
            cell_position = mouse_event.cell_position
        else:
            cell_position = RelativePosition(0.5, 0.5)
        row = mouse_event.position.y + cell_position.y

        # Handle scroll events on the scrollbar
        if mouse_event.event_type == MouseEventType.SCROLL_UP:
            render_info.window._scroll_up()
        elif mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            render_info.window._scroll_down()

        # Mouse drag events
        elif self.dragging and mouse_event.event_type == MouseEventType.MOUSE_MOVE:
            # Add ½ so as to round to nearest integer
            target_scroll = int(
                (row - self.drag_start_offset) / self.track_height * content_height
                + 0.5
            )
            # Use the window's current vertical scroll as it may have changed since the
            # window_render_info was generated
            window = render_info.window
            delta = window.vertical_scroll - target_scroll

            if isinstance(window, Window):
                if delta < 0:
                    func = window._scroll_up
                else:
                    func = window._scroll_down
                for _ in range(abs(delta)):
                    func()
            elif hasattr(window, "scrolling"):
                render_info.window.scrolling = delta

        # Mouse down events
        elif mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            # Scroll up/down one line if clicking on the arrows
            arrows = self.display_arrows()
            if arrows and int(row) == 0:
                offset = -1
            elif arrows and int(row) == render_info.window_height - 1:
                offset = 1
            # Scroll up or down one page if clicking on the background
            elif row < self.thumb_top + 1 or self.thumb_top + 1 + self.thumb_size < row:
                direction = (row < (self.thumb_top + self.thumb_size // 2)) * -2 + 1
                offset = direction * render_info.window_height
            # We are on the scroll button - start a drag event if this is not a
            # repeated mouse event
            elif not repeated:
                self.dragging = True
                self.drag_start_scroll = render_info.vertical_scroll
                self.drag_start_offset = row - self.thumb_top
                # Restrict mouse events to this area
                get_app().mouse_limits = self.write_position
                return None
            # Otherwise this is a click on the centre scroll button - do nothing
            else:
                offset = 0

            if mouse_event.button == MouseButton.LEFT:
                func = None
                if offset < 0:
                    func = render_info.window._scroll_up
                elif offset > 0:
                    func = render_info.window._scroll_down
                if func:
                    # Scroll the window multiple times to scroll by the offset
                    for _ in range(abs(int(offset))):
                        func()
                    # Trigger this mouse event to be repeated
                    self.repeat_task = get_app().create_background_task(
                        self.repeat(mouse_event)
                    )

        # Handle all other mouse events
        else:
            # Stop any repeated tasks
            if self.repeat_task is not None:
                self.repeat_task.cancel()
            # Cancel drags
            self.dragging = False
            get_app().mouse_limits = None
            return NotImplemented

        return None

    def mouse_handler(self, mouse_event: "PtkMouseEvent") -> "NotImplementedOrNone":
        """Type compatible mouse handler."""
        return self._mouse_handler(mouse_event, repeated=False)

    async def repeat(
        self, mouse_event: "PtkMouseEvent", timeout: "float" = 0.1
    ) -> "None":
        """Repeat a mouse event after a timeout."""
        await asyncio.sleep(timeout)
        self._mouse_handler(mouse_event, repeated=True)
        get_app().invalidate()


class NumberedDiffMargin(Margin):
    """Margin that displays the line numbers of a :class:`Window`."""

    style = "class:line-number"

    def get_width(self, get_ui_content: "Callable[[], UIContent]") -> "int":
        """Return the width of the margin."""
        line_count = get_ui_content().line_count
        return len("%s" % line_count) + 2

    def create_margin(
        self, window_render_info: "WindowRenderInfo", width: "int", height: "int"
    ) -> "StyleAndTextTuples":
        """Generate the margin's content."""
        # Get current line number.
        current_lineno = window_render_info.ui_content.cursor_position.y
        # Construct margin.
        result: StyleAndTextTuples = []
        last_lineno = None
        for lineno in window_render_info.displayed_lines:
            # Only display line number if this line is not a continuation of the previous line.
            if lineno != last_lineno:
                if lineno is None:
                    pass
                linestr = str(lineno + 1).rjust(width - 2)
                style = self.style
                if lineno == current_lineno and get_app().layout.has_focus(
                    window_render_info.window
                ):
                    style = f"{style} class:line-number.current"
                result.extend(
                    [
                        (f"{style},edge", "▏"),
                        (style, linestr),
                        (f"{style},edge", "▕"),
                    ]
                )
            last_lineno = lineno
            result.append(("", "\n"))
        return result


class OverflowMargin(Margin):
    """A margin which indicates lines extending beyond the edge of the window."""

    def get_width(self, get_ui_content: "Callable[[], UIContent]") -> "int":
        """Return the width of the margin."""
        return 1

    def create_margin(
        self, window_render_info: "WindowRenderInfo", width: "int", height: "int"
    ) -> "StyleAndTextTuples":
        """Generate the margin's content."""
        from prompt_toolkit.formatted_text.utils import fragment_list_width

        result: "StyleAndTextTuples" = []

        for lineno in window_render_info.displayed_lines:
            line = window_render_info.ui_content.get_line(lineno)
            if (
                fragment_list_width(line) - window_render_info.window.horizontal_scroll
                > window_render_info.window_width + 1
            ):
                result.append(("class:input,overflow", "▹"))
            result.append(("", "\n"))

        return result



"""Defines a command object for use in key-bindings, menus, and the command palette."""

from __future__ import annotations

import logging
import weakref
from inspect import isawaitable, signature
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application import get_app
from prompt_toolkit.filters import to_filter
from prompt_toolkit.key_binding.key_bindings import Binding
from prompt_toolkit.key_binding.key_processor import KeyPressEvent

from euporie.core.key_binding.utils import parse_keys
from euporie.core.keys import Keys

if TYPE_CHECKING:
    from typing import Any, Callable, Coroutine, List, Optional, Sequence, Tuple, Union

    from prompt_toolkit.filters import Filter, FilterOrBool
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        KeyHandlerCallable,
    )

    from euporie.core.widgets.menu import MenuItem

    AnyKey = Union[Tuple[Union[Keys, str], ...], Union[Keys, str]]
    AnyKeys = Union[List[AnyKey], AnyKey]

log = logging.getLogger(__name__)


commands: "dict[str, Command]" = {}


class Command:
    """Wraps a function so it can be used as a key-binding or a menu item."""

    def __init__(
        self,
        handler: "Callable[..., Optional[Coroutine[Any, Any, None]]]",
        *,
        filter: "FilterOrBool" = True,
        hidden: "FilterOrBool" = False,
        name: "Optional[str]" = None,
        title: "Optional[str]" = None,
        menu_title: "Optional[str]" = None,
        description: "Optional[str]" = None,
        toggled: "Optional[Filter]" = None,
        eager: "FilterOrBool" = False,
        is_global: "FilterOrBool" = False,
        save_before: "Callable[[KeyPressEvent], bool]" = (lambda event: True),
        record_in_macro: "FilterOrBool" = True,
    ):
        """Creates a new instance of a command.

        Args:
            handler: The callable to run when the command is triggers
            filter: The condition under which the command is allowed to run
            hidden: The condition under the command is visible to the user
            name: The name of the command, for accessing the command from the registry
            title: The title of the command for display
            menu_title: The title to display in menus if different
            description: The description of the command to explain it's function
            toggled: The toggle state of this command If this command toggles something
            eager: When True, ignore potential longer matches for this key binding
            is_global: Make this a global (always active) binding
            save_before: Determines if the buffer should be saved before running
            record_in_macro: Whether these key bindings should be recorded in macros

        """
        self.handler = handler
        self.filter = to_filter(filter)
        self.hidden = to_filter(hidden)
        if name is None:
            name = handler.__name__.strip("_").replace("_", "-")
        self.name = name
        if title is None:
            title = name.capitalize().replace("-", " ")
        self.title = title
        self.menu_title = menu_title or title
        if description is None:
            # Use the first line of the docstring as the command description
            if handler.__doc__:
                description = (
                    "".join(handler.__doc__.strip().split("\n")).split(".")[0] + "."
                )
            else:
                description = title or name.capitalize()
        self.description = description

        self.toggled = toggled
        self._menu: "Optional[MenuItem]" = None

        self.eager = to_filter(eager)
        self.is_global = to_filter(is_global)
        self.save_before = save_before
        self.record_in_macro = to_filter(record_in_macro)

        self.keys: "list[tuple[Union[str, Keys], ...]]" = []

        self.selected_item = 0
        self.children: "Sequence[MenuItem]" = []

    def run(self) -> "None":
        """Runs the command's handler."""
        if self.filter():
            self.key_handler(
                KeyPressEvent(
                    key_processor_ref=weakref.ref(get_app().key_processor),
                    arg=None,
                    key_sequence=[],
                    previous_key_sequence=[],
                    is_repeat=False,
                )
            )

    @property
    def key_handler(self) -> "KeyHandlerCallable":
        """Returns a key handler for the command."""
        sig = signature(self.handler)

        if sig.parameters:
            # The handler already accepts a `KeyPressEvent` argument
            return cast("KeyHandlerCallable", self.handler)

        def _key_handler(event: "KeyPressEvent") -> "None":
            result = self.handler()
            # If the handler is a coroutine, create an asyncio task.
            if isawaitable(result):
                awaitable = result

                async def bg_task() -> None:
                    result = await awaitable
                    if result != NotImplemented:
                        event.app.invalidate()

                event.app.create_background_task(bg_task())
            elif result != NotImplemented:
                event.app.invalidate()

        return _key_handler

    def bind(self, key_bindings: "KeyBindingsBase", keys: "AnyKeys") -> "None":
        """Add the current commands to a set of key bindings.

        Args:
            key_bindings: The set of key bindings to bind to
            keys: Additional keys to bind to the command

        """
        for key in parse_keys(keys):
            self.keys.append(key)
            key_bindings.bindings.append(
                Binding(
                    key,
                    handler=self.key_handler,
                    filter=self.filter,
                    eager=self.eager,
                    is_global=self.is_global,
                    save_before=self.save_before,
                    record_in_macro=self.record_in_macro,
                )
            )

    def key_str(self) -> "str":
        """Return a string representing the first registered key-binding."""
        from euporie.core.key_binding.utils import format_keys

        if self.keys:
            return format_keys([self.keys[0]])[0]
        return ""

    @property
    def menu_handler(self) -> "Callable[[], None]":
        """Returns a menu handler for the command."""
        if isawaitable(self.handler):

            def _menu_handler() -> "None":
                task = self.handler()
                if task is not None:
                    get_app().create_background_task(task)

            return _menu_handler
        else:
            return cast("Callable[[], None]", self.handler)

    @property
    def menu(self) -> "MenuItem":
        """Returns a menu item for the command."""
        from euporie.core.widgets.menu import MenuItem

        if self._menu is None:
            self._menu = MenuItem.from_command(self)
        return self._menu


def add_cmd(**kwargs: "Any") -> "Callable":
    """Adds a command to the centralized command system."""

    def decorator(handler: "Callable") -> "Callable":
        cmd = Command(handler, **kwargs)
        commands[cmd.name] = cmd
        return handler

    return decorator


def get_cmd(name: "str") -> "Command":
    """Get a command from the centralized command system by name.

    Args:
        name: The name of the command to retrieve

    Returns:
        The requested command object

    Raises:
        KeyError: Raised if the named command is not found

    """
    try:
        return commands[name]
    except KeyError as e:
        raise KeyError("Unknown command: %r" % name) from e



"""Contains functions which convert data to png format."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from euporie.core.convert.base import register
from euporie.core.convert.formats.common import base64_to_bytes_py, imagemagick_convert
from euporie.core.convert.utils import commands_exist, have_modules

if TYPE_CHECKING:
    from typing import Optional

    from PIL import Image
    from upath import UPath


register(
    from_="base64-png",
    to="png",
)(base64_to_bytes_py)


def latex_to_png_py_ipython(
    backend: "str",
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "bytes":
    """Converts LaTeX data to PNG bytes with :py:mod:`IPython` & :py:mod:`matplotlib`."""
    from IPython.lib.latextools import latex_to_png

    return (
        latex_to_png(
            data, encode=False, backend=backend, **({"color": fg} if fg else {})
        )
        or b"error"
    )


register(
    from_="latex",
    to="png",
    filter_=commands_exist("dvipng") & have_modules("IPython.lib.latextools"),
)(partial(latex_to_png_py_ipython, "dvipng"))


register(
    from_="latex",
    to="png",
    filter_=have_modules("IPython.lib.latextools", "matplotlib"),
)(partial(latex_to_png_py_ipython, "matplotlib"))


register(
    from_=("svg", "jpeg", "pdf"),
    to="png",
    filter_=commands_exist("convert", "magick"),
)(partial(imagemagick_convert, "PNG"))


@register(
    from_="pil",
    to="png",
    filter_=have_modules("PIL"),
)
def pil_to_png_py_pil(
    data: "Image",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
) -> "bytes":
    """Convert a pillow image to sixels :py:mod:`teimpy`."""
    import io

    with io.BytesIO() as output:
        data.save(output, format="PNG")
        contents = output.getvalue()
    return contents


@register(
    from_="svg",
    to="png",
    filter_=have_modules("cairosvg"),
)
def svg_to_png_py_cairosvg(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
) -> "str":
    """Convert SVG to PNG using :py:mod:`cairosvg`."""
    import cairosvg

    return cairosvg.surface.PNGSurface.convert(data, write_to=None)



"""Style related functions."""

from __future__ import annotations

import logging
from colorsys import hls_to_rgb, rgb_to_hls
from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.styles.base import DEFAULT_ATTRS, BaseStyle
from prompt_toolkit.styles.defaults import default_ui_style
from prompt_toolkit.styles.style import Style, _parse_style_str, merge_styles

if TYPE_CHECKING:
    from typing import Any, Hashable, Optional

    from prompt_toolkit.styles.base import Attrs
    from prompt_toolkit.styles.style import _MergedStyle

log = logging.getLogger(__name__)


KERNEL_STATUS_REPR = {
    "stopped": "⨂",
    "starting": "◍",
    "idle": "○",
    "busy": "●",
    "error": "☹",
}


DEFAULT_COLORS = {
    "bg": "#232627",
    "fg": "#fcfcfc",
    "ansiblack": "#000000",
    "ansired": "#cc0403",
    "ansigreen": "#19cb00",
    "ansiyellow": "#cecb00",
    "ansiblue": "#0d73cc",
    "ansipurple": "#cb1ed1",
    "ansimagenta": "#cb1ed1",
    "ansicyan": "#0dcdcd",
    "ansiwhite": "#dddddd",
    "ansibrightblack": "#767676",
    "ansigray": "#767676",
    "ansibrightred": "#f2201f",
    "ansibrightgreen": "#23fd00",
    "ansibrightyellow": "#fffd00",
    "ansibrightblue": "#1a8fff",
    "ansibrightpurple": "#fd28ff",
    "ansibrightmagenta": "#fd28ff",
    "ansibrightcyan": "#14ffff",
    "ansibrightwhite": "#ffffff",
}


MIME_STYLE = [
    ("mime.stream.stderr", "fg:ansired"),
]

HTML_STYLE = [
    ("html h1", "bold underline"),
    ("html h1 border", "fg:ansiyellow nounderline"),
    ("html h2", "bold"),
    ("html h2 border", "fg:grey nobold"),
    ("html h3", "bold"),
    ("html h4", "bold italic"),
    ("html h5", "underline"),
    ("html h6", "italic"),
    ("html code", "bg:#333"),
    ("html kbd", "fg:#fff bg:#333"),
    ("html samp", "fg:#fff bg:#333"),
    ("html b", "bold"),
    ("html strong", "bold"),
    ("html i", "italic"),
    ("html em", "italic"),
    ("html cite", "italic"),
    ("html dfn", "italic"),
    ("html var", "italic"),
    ("html u", "underline"),
    ("html ins", "underline"),
    ("html s", "strike"),
    ("html del", "strike"),
    ("html mark", "fg:black bg:ansiyellow"),
    ("html hr", "fg:ansired"),
    ("html ul bullet", "fg:ansiyellow"),
    ("html ol bullet", "fg:ansicyan"),
    ("html blockquote", "fg:ansipurple"),
    ("html blockquote margin", "fg:grey"),
    ("html th", "bold"),
    ("html a", "underline fg:ansibrightblue"),
    ("html img", "bg:cyan fg:black"),
    ("html img border", "fg:cyan bg:default"),
    ("html caption", "italic"),
    ("html summary", "bold"),
]


LOG_STYLE = [
    ("log.level.nonset", "fg:ansigray"),
    ("log.level.debug", "fg:ansigreen"),
    ("log.level.info", "fg:ansiblue"),
    ("log.level.warning", "fg:ansiyellow"),
    ("log.level.error", "fg:ansired"),
    ("log.level.critical", "fg:ansiwhite bg:ansired bold"),
]

IPYWIDGET_STYLE = [
    ("ipywidget", "bg:default"),
    ("ipywidget focused", ""),
    ("ipywidget button face success", "bg:ansigreen"),
    ("ipywidget button face", "fg:black bg:#d4d0c8"),
    ("ipywidget button face success", "bg:ansigreen"),
    ("ipywidget button face info", "bg:ansicyan"),
    ("ipywidget button face warning", "bg:ansiyellow"),
    ("ipywidget button face danger", "bg:ansired"),
    ("ipywidget button border right", "fg:#606060"),
    ("ipywidget button border bottom", "fg:#606060"),
    ("ipywidget button border left", "fg:#ffffff"),
    ("ipywidget button border top", "fg:#ffffff"),
    ("ipywidget button face disabled", "fg:#888888"),
    ("ipywidget selection border right", "fg:#ffffff"),
    ("ipywidget selection border bottom", "fg:#ffffff"),
    ("ipywidget selection border left", "fg:#606060"),
    ("ipywidget selection border top", "fg:#606060"),
    ("ipywidget success border left", "fg:ansibrightgreen"),
    ("ipywidget success border top", "fg:ansibrightgreen"),
    ("ipywidget info border left", "fg:ansibrightcyan"),
    ("ipywidget info border top", "fg:ansibrightcyan"),
    ("ipywidget warning border left", "fg:ansibrightyellow"),
    ("ipywidget warning border top", "fg:ansibrightyellow"),
    ("ipywidget danger border left", "fg:ansibrightred"),
    ("ipywidget danger border top", "fg:ansibrightred"),
    ("ipywidget selection border left", "fg:#606060"),
    ("ipywidget selection border top", "fg:#606060"),
    ("ipywidget selection success border right", "fg:ansibrightgreen"),
    ("ipywidget selection success border bottom", "fg:ansibrightgreen"),
    ("ipywidget selection success border right", "fg:ansibrightgreen"),
    ("ipywidget selection success border bottom", "fg:ansibrightgreen"),
    ("ipywidget selection info border right", "fg:ansibrightcyan"),
    ("ipywidget selection info border bottom", "fg:ansibrightcyan"),
    ("ipywidget selection danger border right", "fg:ansibrightred"),
    ("ipywidget selection danger border bottom", "fg:ansibrightred"),
    ("ipywidget text text-area", "fg:black bg:white"),
    ("ipywidget text placeholder", "fg:#AAAAAA bg:white"),
    ("ipywidget text border right", "fg:#E9E7E3"),
    ("ipywidget text border top", "fg:#606060"),
    ("ipywidget text border bottom", "fg:#E9E7E3"),
    ("ipywidget text border left", "fg:#606060"),
    ("ipywidget text border top invalid", "fg:ansidarkred"),
    ("ipywidget text border right invalid", "fg:ansired"),
    ("ipywidget text border bottom invalid", "fg:ansired"),
    ("ipywidget text border left invalid", "fg:ansidarkred"),
    ("ipywidget slider track selection", "fg:ansiblue"),
    ("ipywidget slider handle focused", "fg:white"),
    ("ipywidget slider handle selection", ""),
    ("ipywidget slider handle selection focused", "fg:ansiblue"),
    ("ipywidget progress", "fg:ansidarkblue bg:#d4d0c8"),
    ("ipywidget progress success", "fg:ansigreen"),
    ("ipywidget progress info", "fg:ansicyan"),
    ("ipywidget progress warning", "fg:ansiyellow"),
    ("ipywidget progress danger", "fg:ansired"),
    ("ipywidget progress border", "bg:default"),
    ("ipywidget progress border top", "fg:#606060"),
    ("ipywidget progress border right", "fg:#E9E7E3"),
    ("ipywidget progress border bottom", "fg:#E9E7E3"),
    ("ipywidget progress border left", "fg:#606060"),
    ("ipywidget dropdown dropdown.menu", "fg:#000000 bg:#ffffff"),
    ("ipywidget dropdown dropdown.menu hovered", "fg:#ffffff bg:ansiblue"),
    ("ipywidget checkbox selection", "fg:default"),
    ("ipywidget checkbox prefix selection", "fg:ansiblue"),
    ("ipywidget checkbox disabled", "fg:#888888"),
    ("ipywidget checkbox prefix selection disabled", "fg:#888888"),
    ("ipywidget valid prefix", "fg:ansired"),
    ("ipywidget valid prefix selection", "fg:ansigreen"),
    ("ipywidget radio-buttons selection", "fg:default"),
    ("ipywidget radio-buttons selection disabled", "fg:#888888"),
    ("ipywidget radio-buttons prefix selection", "fg:ansiblue"),
    ("ipywidget select face", "fg:black bg:white"),
    ("ipywidget select face selection", "fg:white bg:ansiblue"),
    ("ipywidget select face hovered", "fg:black bg:#f0f0f0"),
    ("ipywidget select face hovered selection", "fg:white bg:ansiblue"),
    ("ipywidget select border top", "fg:#606060"),
    ("ipywidget select border right", "fg:#E9E7E3"),
    ("ipywidget select border bottom", "fg:#E9E7E3"),
    ("ipywidget select border left", "fg:#606060"),
    ("ipywidget select disabled", "fg:#888888"),
    ("ipywidget select selection disabled", "fg:#888888"),
    ("ipywidget inset border right", "fg:#E9E7E3"),
    ("ipywidget inset border top", "fg:#606060"),
    ("ipywidget inset border bottom", "fg:#E9E7E3"),
    ("ipywidget inset border left", "fg:#606060"),
    ("ipywidget swatch border right", "fg:#E9E7E3"),
    ("ipywidget swatch border top", "fg:#606060"),
    ("ipywidget swatch border bottom", "fg:#E9E7E3"),
    ("ipywidget swatch border left", "fg:#606060"),
    ("ipywidget tabbed-split tab-bar tab active border top default", "fg:ansiblue"),
    ("ipywidget tabbed-split tab-bar tab active border top success", "fg:ansigreen"),
    ("ipywidget tabbed-split tab-bar tab active border top info", "fg:ansicyan"),
    ("ipywidget tabbed-split tab-bar tab active border top warning", "fg:ansiyellow"),
    ("ipywidget tabbed-split tab-bar tab active border top danger", "fg:ansired"),
    ("ipywidget accordion border success", "fg:ansigreen"),
    ("ipywidget accordion border info", "fg:ansicyan"),
    ("ipywidget accordion border warning", "fg:ansiyellow"),
    ("ipywidget accordion border danger", "fg:ansired"),
    ("ipywidget accordion selection", "fg:ansiblue"),
]


class ColorPaletteColor:
    """A representation of a color with adjustment methods."""

    _cache: "SimpleCache[tuple[str, float, float, float, bool], ColorPaletteColor]" = (
        SimpleCache()
    )

    def __init__(self, base: "str", _base_override: str = "") -> "None":
        """Creates a new color."""
        self.base_hex = DEFAULT_COLORS.get(base, base)
        self.base = _base_override or base

        color = self.base_hex.lstrip("#")
        self.red, self.green, self.blue = (
            int(color[0:2], 16) / 255,
            int(color[2:4], 16) / 255,
            int(color[4:6], 16) / 255,
        )

        self.hue, self.brightness, self.saturation = rgb_to_hls(
            self.red, self.green, self.blue
        )

        self.is_light = self.brightness > 0.5

    def _adjust_abs(
        self, hue: "float" = 0.0, brightness: "float" = 0.0, saturation: "float" = 0.0
    ) -> "ColorPaletteColor":
        hue = max(min(1, self.hue + hue), 0)
        brightness = max(min(1, self.brightness + brightness), 0)
        saturation = max(min(1, self.saturation + saturation), 0)

        r, g, b = hls_to_rgb(hue, brightness, saturation)
        new_color = f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"
        return ColorPaletteColor(new_color)

    def _adjust_rel(
        self, hue: "float" = 0.0, brightness: "float" = 0.0, saturation: "float" = 0.0
    ) -> "ColorPaletteColor":
        hue = min(max(0, hue), 1)
        brightness = min(max(-1, brightness), 1)
        saturation = min(max(-1, saturation), 1)

        new_hue = self.hue + (self.hue * (hue < 0) + (1 - self.hue) * (hue > 0)) * hue

        new_brightness = (
            self.brightness
            + (
                self.brightness * (brightness < 0)
                + (1 - self.brightness) * (brightness > 0)
            )
            * brightness
        )

        new_saturation = (
            self.saturation
            + (
                self.saturation * (saturation < 0)
                + (1 - self.saturation) * (saturation > 0)
            )
            * saturation
        )

        r, g, b = hls_to_rgb(new_hue, new_brightness, new_saturation)
        new_color = f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"
        return ColorPaletteColor(new_color)

    def _adjust(
        self,
        hue: "float" = 0.0,
        brightness: "float" = 0.0,
        saturation: "float" = 0.0,
        rel: "bool" = True,
    ) -> "ColorPaletteColor":
        """Performs a relative of absolute color adjustment."""
        if rel:
            return self._adjust_rel(hue, brightness, saturation)
        else:
            return self._adjust_abs(hue, brightness, saturation)

    def adjust(
        self,
        hue: "float" = 0.0,
        brightness: "float" = 0.0,
        saturation: "float" = 0.0,
        rel: "bool" = True,
    ) -> "ColorPaletteColor":
        """Adjust the hue, saturation, or brightness of the color."""
        key = (self.base_hex, hue, brightness, saturation, rel)
        return self._cache.get(
            key, partial(self._adjust, hue, brightness, saturation, rel)
        )

    def lighter(self, amount: "float", rel: "bool" = True) -> "ColorPaletteColor":
        """Makes the color lighter."""
        return self.adjust(brightness=amount, rel=rel)

    def darker(self, amount: "float", rel: "bool" = True) -> "ColorPaletteColor":
        """Makes the color darker."""
        return self.adjust(brightness=-amount, rel=rel)

    def more(self, amount: "float", rel: "bool" = True) -> "ColorPaletteColor":
        """Makes bright colors darker and dark colors brighter."""
        if self.is_light:
            amount *= -1
        return self.adjust(brightness=amount, rel=rel)

    def less(self, amount: "float", rel: "bool" = True) -> "ColorPaletteColor":
        """Makes bright colors brighter and dark colors darker."""
        if self.is_light:
            amount *= -1
        return self.adjust(brightness=-amount, rel=rel)

    def towards(
        self, other: "ColorPaletteColor", amount: "float"
    ) -> "ColorPaletteColor":
        """Interpolate between two colors."""
        amount = min(max(0, amount), 1)
        r = (other.red - self.red) * amount + self.red
        g = (other.green - self.green) * amount + self.green
        b = (other.blue - self.blue) * amount + self.blue
        new_color = f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"
        return ColorPaletteColor(new_color)

    def __repr__(self) -> "str":
        """Returns a string representation of the color."""
        return self.base


class ColorPalette:
    """Defines a collection of colors."""

    def __init__(self):
        """Creates a new color-palette."""
        self.colors = {}

    def add_color(
        self, name: "str", base: "str", _base_override: str = ""
    ) -> "ColorPalette":
        """Adds a color to the palette."""
        self.colors[name] = ColorPaletteColor(base, _base_override)
        return self

    def __getattr__(self, name: "str") -> "Any":
        """Enables access of palette colors via dotted attributes.

        Args:
            name: The name of the attribute to access.

        Returns:
            The color-palette color.

        """
        return self.colors.get(name)


def build_style(
    cp: "ColorPalette",
    have_term_colors: "bool" = True,
) -> "Style":
    """Create an application style based on the given color palette."""
    style_dict = {
        # The default style is merged at this point so full styles can be
        # overridden. For example, this allows us to switch off the underline
        # status of cursor-line.
        **dict(default_ui_style().style_rules),
        "default": f"fg:{cp.bg.base} bg:{cp.bg.base}",
        # Logo
        "logo": "fg:#dd0000",
        # Pattern
        "pattern": f"fg:{cp.bg.more(0.075)}",
        # Chrome
        "chrome": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "tab-padding": f"fg:{cp.bg.more(0.2)} bg:{cp.bg.base}",
        # Statusbar
        "status": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "status.field": f"fg:{cp.fg.more(0.1)} bg:{cp.bg.more(0.1)}",
        # Menus & Menu bar
        "menu": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "menu disabled": f"fg:{cp.fg.more(0.05).towards(cp.bg, 0.75)}",
        "menu shortcut": f"fg:{cp.fg.more(0.4)}",
        "menu shortcut disabled": f"fg:{cp.fg.more(0.4).towards(cp.bg, 0.5)}",
        "menu prefix": f"fg:{cp.fg.more(0.2)}",
        "menu prefix disabled": f"fg:{cp.fg.more(0.2).towards(cp.bg, 0.5)}",
        "menu selection": f"fg:{cp.hl.more(1)} bg:{cp.hl}",
        "menu selection shortcut": f"fg:{cp.hl.more(1).more(0.05)} bg:{cp.hl}",
        "menu selection prefix": f"fg:{cp.hl.more(1).more(0.05)} bg:{cp.hl}",
        "menu border": f"fg:{cp.bg.more(0.15)} bg:{cp.bg.more(0.05)}",
        "menu border selection": f"bg:{cp.hl}",
        # Tab bar
        "app-tab-bar": f"bg:{cp.bg.less(0.15)}",
        "app-tab-bar border": f"fg:{cp.bg.more(0.1)}",
        "app-tab-bar tab inactive": f"fg:{cp.fg.more(0.5)}",
        "app-tab-bar tab inactive border": f"fg:{cp.bg.more(0.15)}",
        "app-tab-bar tab active": "bold fg:default bg:default",
        "app-tab-bar tab active close": "fg:darkred",
        "app-tab-bar tab active border top": f"fg:{cp.hl} bg:{cp.bg.less(0.15)}",
        # Buffer
        "line-number": f"fg:{cp.fg.more(0.5)} bg:{cp.bg.more(0.05)}",
        "line-number.current": f"bold orange bg:{cp.bg.more(0.1)}",
        "line-number edge": f"fg:{cp.bg.darker(0.1)}",
        "line-number.current edge": f"fg:{cp.bg.darker(0.1)}",
        "cursor-line": f"bg:{cp.bg.more(0.05)}",
        "cursor-line search": f"bg:{cp.bg.more(0.02)}",
        "cursor-line search.current": f"bg:{cp.bg.more(0.02)}",
        "cursor-line incsearch": "bg:ansibrightyellow",
        "cursor-line incsearch.current": "bg:ansibrightgreen",
        "matching-bracket.cursor": "fg:yellow bold",
        "matching-bracket.other": "fg:yellow bold",
        "trailing-whitespace": f"fg:{cp.fg.more(0.66)}",
        "tab": f"fg:{cp.fg.more(0.66)}",
        # Search
        "search": f"bg:{cp.bg.more(0.05)}",
        "search.current": f"bg:{cp.bg.more(0.05)}",
        "incsearch": "bg:ansibrightyellow",
        "incsearch.current": "bg:ansibrightgreen",
        "search-toolbar": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "search-toolbar.title": f"fg:{cp.fg.more(0.1)} bg:{cp.bg.more(0.1)}",
        # Inputs
        "kernel-input": f"fg:default bg:{cp.bg.more(0.02)}",
        # Cells
        "cell.border": f"fg:{cp.bg.more(0.25)}",
        "cell.border.selected": f"fg:{cp.hl.more(0.2)}",
        "cell.border.edit": "fg:ansibrightgreen",
        "cell.output": "fg:default bg:default",
        "cell.input.prompt": "fg:blue",
        "cell.output.prompt": "fg:red",
        "cell show outputs": "bg:#888",
        "cell show inputs": "bg:#888",
        # Scrollbars
        "scrollbar": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.15)}",
        "scrollbar.background": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.15)}",
        "scrollbar.arrow": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.15)}",
        "scrollbar.start": "",
        # "scrollbar.start": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.25)}",
        "scrollbar.button": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.75)}",
        "scrollbar.end": f"fg:{cp.bg.more(0.15)} bg:{cp.bg.more(0.75)}",
        # Overflow margin
        "overflow": f"fg:{cp.fg.more(0.5)}",
        # Dialogs
        "dialog title": f"fg:white bg:{cp.hl.darker(0.25)} bold",
        "dialog title border": "fg:ansired",
        "dialog": f"fg:{cp.fg.base} bg:{cp.bg.darker(0.1)}",
        "dialog text-area": f"bg:{cp.bg.lighter(0.05)}",
        "dialog input text text-area": f"fg:default bg:{cp.bg.less(0.1)}",
        "dialog text-area last-line": "nounderline",
        "dialog border": f"fg:{cp.bg.darker(0.1).more(0.1)}",
        # Horizontals rule
        "hr": "fg:ansired",
        # Completions menu
        "completion-menu.completion.keyword": "fg:#d700af",
        "completion-menu.completion.current.keyword": "fg:#fff bg:#d700ff",
        "completion-menu.completion.function": "fg:#005faf",
        "completion-menu.completion.current.function": "fg:#fff bg:#005fff",
        "completion-menu.completion.class": "fg:#008700",
        "completion-menu.completion.current.class": "fg:#fff bg:#00af00",
        "completion-menu.completion.statement": "fg:#5f0000",
        "completion-menu.completion.current.statement": "fg:#fff bg:#5f0000",
        "completion-menu.completion.instance": "fg:#d75f00",
        "completion-menu.completion.current.instance": "fg:#fff bg:#d78700",
        "completion-menu.completion.module": "fg:#d70000",
        "completion-menu.completion.current.module": "fg:#fff bg:#d70000",
        # Log
        "log.level.nonset": "fg:grey",
        "log.level.debug": "fg:green",
        "log.level.info": "fg:blue",
        "log.level.warning": "fg:yellow",
        "log.level.error": "fg:red",
        "log.level.critical": "fg:red bold",
        "log.ref": "fg:grey",
        "log.date": "fg:#00875f",
        # Shortcuts
        "shortcuts.group": f"bg:{cp.bg.more(0.4)} bold underline",
        # "shortcuts.row": f"bg:{cp.bg.base} nobold",
        "shortcuts.row alt": f"bg:{cp.bg.more(0.1)}",
        "shortcuts.row key": "bold",
        # Palette
        "palette.item": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "palette.item.alt": f"bg:{cp.bg.more(0.15)}",
        "palette.item.selected": f"fg:{cp.hl.more(1)} bg:{cp.hl}",
        # Pager
        "pager": f"bg:{cp.bg.more(0.05)}",
        "pager.border": f"fg:{cp.bg.more(0.45)}",
        # Markdown
        "html code": f"bg:{cp.bg.more(0.15)}",
        "html code block": f"bg:{cp.bg.less(0.2)}",
        "html pre border": f"fg:{cp.bg.more(0.25)}",
        "html table border": f"fg:{cp.bg.more(0.75)}",
        # Drop-shadow
        "drop-shadow.inner": f"fg:{cp.bg.darker(0.15)}",
        "drop-shadow.outer": f"fg:{cp.bg.darker(0.1)} bg:{cp.bg.darker(0.025)}",
        # Tabbed split
        "tabbed-split tab-bar tab inactive": f"fg:{cp.bg.more(0.1)}",
        "tabbed-split tab-bar tab inactive border": f"fg:{cp.bg.more(0.15)}",
        "tabbed-split tab-bar tab active": f"bold fg:{cp.fg}",
        "tabbed-split tab-bar tab active close": "fg:darkred",
        "tabbed-split border": f"fg:{cp.bg.more(0.2)}",
        # Ipywidgets
        "ipywidget focused": f"bg:{cp.bg.more(0.05)}",
        "ipywidget slider track": f"fg:{cp.fg.darker(0.5)}",
        "ipywidget slider arrow": f"fg:{cp.fg.darker(0.25)}",
        "ipywidget slider handle": f"fg:{cp.fg.darker(0.25)}",
        "ipywidget accordion border default": f"fg:{cp.bg.more(0.2)}",
        # Input widgets
        "text-area selected": "noreverse",
        "text-area selected focused": "reverse",
        # "input focused": f"bg:{cp.bg.more(0.025)}",
        # "input button face": f"fg:default bg:{cp.bg.more(0.05)}",
        # "input button face focused": f"fg:default bg:{cp.hl.darker(0.75)}",
        # "input button border top": "fg:#ffffff",
        # "input button border right": "fg:#606060",
        # "input button border bottom": "fg:#606060",
        # "input button border left": "fg:#ffffff",
        "input button border top": f"fg:{cp.bg.lighter(0.5)}",
        "input button border right": f"fg:{cp.bg.darker(0.25)}",
        "input button border bottom": f"fg:{cp.bg.darker(0.25)}",
        "input button border left": f"fg:{cp.bg.lighter(0.5)}",
        "input button border top focused": f"fg:{cp.hl.lighter(0.5)}",
        "input button border right focused": f"fg:{cp.hl.darker(0.5)}",
        "input button border bottom focused": f"fg:{cp.hl.darker(0.5)}",
        "input button border left focused": f"fg:{cp.hl.lighter(0.5)}",
        "input selection border right": "fg:#ffffff",
        "input selection border bottom": "fg:#ffffff",
        "input selection border top": "fg:#606060",
        "input selection border left": "fg:#606060",
        "input selection focused border right": f"fg:{cp.hl}",
        "input selection focused border bottom": f"fg:{cp.hl}",
        "input selection focused border top": f"fg:{cp.hl.darker(0.5)}",
        "input selection focused border left": f"fg:{cp.hl.darker(0.5)}",
        "input text placeholder": f"fg:{cp.fg.more(0.6)}",
        # "input text border top": "fg:#606060",
        # "input text border right": "fg:#E9E7E3",
        # "input text border bottom": "fg:#E9E7E3",
        # "input text border left": "fg:#606060",
        "input text text-area": f"bg:{cp.bg.lighter(0.1)}",
        "input text border top": f"fg:{cp.bg.darker(0.5)}",
        "input text border right": f"fg:{cp.bg.lighter(0.25)}",
        "input text border bottom": f"fg:{cp.bg.lighter(0.25)}",
        "input text border left": f"fg:{cp.bg.darker(0.5)}",
        "input text border top focused": f"fg:{cp.hl.darker(0.5)}",
        "input text border right focused": f"fg:{cp.hl.lighter(0.5)}",
        "input text border bottom focused": f"fg:{cp.hl.lighter(0.5)}",
        "input text border left focused": f"fg:{cp.hl.darker(0.5)}",
        "input text border top invalid": "fg:ansidarkred",
        "input text border right invalid": "fg:ansired",
        "input text border bottom invalid": "fg:ansired",
        "input text border left invalid": "fg:ansidarkred",
        "radio-buttons prefix selection focused": f"fg:{cp.hl}",
        # Dataframes
        "html table dataframe border": f"fg:{cp.bg.more(0.25)} bg:{cp.bg.more(0.05)}",
        "html table dataframe th": f"bg:{cp.bg.more(0.1)}",
    }

    return Style.from_dict(style_dict)


class ShadowStyle(BaseStyle):
    """Wraps a :py:class:`BaseStyle`, and generates opaque shadow styles for each rule."""

    def __init__(
        self,
        style: "BaseStyle",
        color_palette: "ColorPalette",
        opacity: "float" = 0.5,
        color: "Optional[ColorPaletteColor]" = None,
    ) -> "None":
        """The constructor accepts a style."""
        self.cp = color_palette
        self.color = color or color_palette.bg.darker(0.5)
        self.opacity = opacity
        self.base_style = style
        self._style: "SimpleCache[Hashable, _MergedStyle]" = SimpleCache(maxsize=1)

    @property
    def _merged_style(self) -> "_MergedStyle":
        """The `Style` object that has the other styles merged together."""

        def get() -> "_MergedStyle":
            return merge_styles([self.base_style, Style(self.shadow_rules)])

        return self._style.get(self.base_style.invalidation_hash(), get)

    @property
    def shadow_rules(self) -> "list[tuple[str, str]]":
        """Generate new shadow rules."""
        color = self.color
        opacity = self.opacity

        default_fg = self.cp.fg.towards(self.color, opacity)
        default_bg = self.cp.bg.towards(self.color, opacity)

        new_rules = [("shadow", f"fg:{default_fg} bg:{default_bg}")]
        for rule, style_str in self.base_style.style_rules:
            style_attrs = _parse_style_str(style_str)
            if "shadow" in rule.split():
                continue
            new_style_str = ""
            old_fg = style_attrs.color
            old_bg = style_attrs.bgcolor
            new_fg = (
                ColorPaletteColor(old_fg).towards(color, opacity)
                if old_fg and old_fg not in ("default", "ansidefault")
                else default_fg
            )
            new_bg = (
                ColorPaletteColor(old_bg).towards(color, opacity)
                if old_bg and old_bg not in ("default", "ansidefault")
                else default_bg
            )
            new_style_str += f"fg:{new_fg} bg:{new_bg}"
            new_rules.append((f"{rule} shadow", new_style_str))
        return new_rules

    @property
    def style_rules(self) -> "list[tuple[str, str]]":
        """The list of style rules, used to create this style."""
        return self._merged_style.style_rules

    def get_attrs_for_style_str(
        self, style_str: "str", default: "Attrs" = DEFAULT_ATTRS
    ) -> "Attrs":
        """Return :class:`.Attrs` for the given style string."""
        return self._merged_style.get_attrs_for_style_str(style_str, default)

    def invalidation_hash(self) -> "Hashable":
        """Invalidation hash for the style, which redraws everything on a change."""
        return self.base_style.invalidation_hash()



"""Contains functions which convert data to html format."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from euporie.core.convert.base import register
from euporie.core.convert.utils import have_modules

if TYPE_CHECKING:
    from typing import Optional

    from upath import UPath

log = logging.getLogger(__name__)

# Check for markdown-it-py
markdown_parser: "Optional[MarkdownIt]" = None

try:
    from markdown_it import MarkdownIt
except ModuleNotFoundError:
    pass
else:

    class MarkdownParser(MarkdownIt):
        """Subclass the markdown parser to allow ``file:`` URIs."""

        def validateLink(self, url: "str") -> "bool":
            """Allows all link URIs."""
            return True

    markdown_parser = (
        MarkdownParser().enable("linkify").enable("table").enable("strikethrough")
    )

# Check for markdown-it-py plugins
try:
    import mdit_py_plugins  # noqa F401
except ModuleNotFoundError:
    pass
else:
    from mdit_py_plugins.amsmath import amsmath_plugin
    from mdit_py_plugins.dollarmath.index import dollarmath_plugin
    from mdit_py_plugins.texmath.index import texmath_plugin

    # from mdit_py_plugins.tasklists import tasklists_plugin

    if markdown_parser is not None:
        markdown_parser.use(texmath_plugin)
        markdown_parser.use(dollarmath_plugin)
        markdown_parser.use(amsmath_plugin)
        # markdown_parser.use(tasklists_plugin)


@register(from_="markdown", to="html", filter_=have_modules("markdown_it"))
def markdown_to_html_markdown_it(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Convert markdown to HTML using :py:mod:`markdownit_py`."""
    assert markdown_parser is not None
    # Give markdown tables borders
    html = "<style>table{border-width:1}th{border-width:3;font-weight:bold}</style>"
    html += markdown_parser.render(data)
    return html



"""Contains functions which can be used to convert data to multiple formats."""

from __future__ import annotations

import base64
import logging
from typing import TYPE_CHECKING

from euporie.core.app import get_app
from euporie.core.convert.utils import call_subproc

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from upath import UPath

log = logging.getLogger(__name__)


def base64_to_bytes_py(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "bytes":
    """Converts base64 encoded data to bytes."""
    return base64.b64decode(data)


def imagemagick_convert(
    output_format: "str",
    data: "Union[str, bytes]",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "Union[str, bytes]":
    """Converts image data to PNG bytes using ``imagemagick``."""
    cmd: "list[Any]" = ["convert"]
    if cols is not None:
        px, _ = get_app().term_info.cell_size_px
        cmd += ["-geometry", f"{int(cols * px)}"]
    bg = bg or get_app().color_palette.bg.base_hex
    if bg is not None:
        cmd += ["-background", bg]
    cmd += ["-", f"{output_format}:-"]
    result: "Union[bytes, str]" = call_subproc(data, cmd)
    if output_format in {"sixel", "svg"} and isinstance(result, bytes):
        result = result.decode()
    return result


def chafa_convert(
    output_format: "str",
    data: "Union[bytes, str]",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "Union[str, bytes]":
    """Converts image data to ANSI text using :command:`chafa`."""
    cmd: "list[Any]" = ["chafa", f"--format={output_format}"]
    if cols is not None and rows is not None:
        cmd += [f"--size={cols}x{rows}"]
    cmd += ["--stretch", "-"]
    return call_subproc(data, cmd).decode()



"""Contains a container for the cell output area."""

from __future__ import annotations

import logging
from pathlib import PurePath
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    HSplit,
)
from prompt_toolkit.widgets import Box

from euporie.core.app import get_app
from euporie.core.commands import add_cmd
from euporie.core.convert.base import BASE64_FORMATS, MIME_FORMATS, find_route
from euporie.core.filters import pager_has_focus
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.widgets.cell_outputs import CellOutput, CellOutputDataElement
from euporie.core.widgets.decor import Line
from euporie.core.widgets.display import Display

if TYPE_CHECKING:
    from typing import Any, Optional

    from prompt_toolkit.layout.dimension import AnyDimension

    from euporie.core.widgets.cell_outputs import CellOutputElement, OutputParent

log = logging.getLogger(__name__)


PagerState = NamedTuple(
    "PagerState",
    [("code", str), ("cursor_pos", int), ("response", dict)],
)


class PagerOutputDataElement(CellOutputDataElement):
    """A cell output element which display data."""

    def __init__(
        self,
        mime: "str",
        data: "dict[str, Any]",
        metadata: "dict",
        parent: "Optional[OutputParent]",
    ) -> "None":
        """Create a new data output element instance.

        Args:
            mime: The mime-type of the data to display
            data: The data to display
            metadata: Any metadata relating to the data
            parent: The parent container the output-element is attached to
        """
        # Get internal format
        format_ = "ansi"
        mime_path = PurePath(mime)
        for mime_type, data_format in MIME_FORMATS.items():
            if mime_path.match(mime_type):
                if data_format in BASE64_FORMATS:
                    data_format = f"base64-{data_format}"
                if find_route(data_format, "formatted_text") is not None:
                    format_ = data_format
                    break

        self.container = Display(
            data=data,
            format_=format_,
            focusable=True,
            focus_on_click=True,
            wrap_lines=True,
            always_hide_cursor=True,
            style="class:pager",
            scrollbar_autohide=False,
            dont_extend_height=False,
        )


class PagerOutput(CellOutput):
    """Display pager output."""

    def make_element(self, mime: "str") -> "CellOutputElement":
        """Creates a container for the pager output mime-type if it doesn't exist.

        Args:
            mime: The mime-type to make the container for

        Returns:
            A :class:`PagerOutputDataElement` container for the currently selected mime-type.
        """
        return PagerOutputDataElement(
            mime=self.selected_mime,
            data=self.data[self.selected_mime],
            metadata=self.json.get("metadata", {}).get(self.selected_mime, {}),
            parent=self.parent,
        )


class Pager:
    """Interactive help pager.

    A pager which displays information at the bottom of a tab.

    """

    def __init__(self, height: "Optional[AnyDimension]" = None) -> "None":
        """Create a new page instance."""
        self._state: "Optional[PagerState]" = None
        self.visible = Condition(
            lambda: self.state is not None and bool(self.state.response.get("found"))
        )
        self.output = PagerOutput({}, None)

        self.container = ConditionalContainer(
            HSplit(
                [
                    Line(
                        height=1,
                        collapse=False,
                        style="class:pager.border",
                    ),
                    Box(
                        DynamicContainer(lambda: self.output),
                        padding=0,
                        padding_left=1,
                    ),
                ],
                height=height or (lambda: get_app().output.get_size().rows // 3),
                style="class:pager",
                key_bindings=load_registered_bindings(
                    "euporie.core.widgets.pager.Pager"
                ),
            ),
            filter=self.visible,
        )

    def focus(self) -> "None":
        """Focus the pager."""
        get_app().layout.focus(self)

    def hide(self) -> "None":
        """Clear and hide the pager."""
        if self.visible():
            self.state = None
            # Focus previous control if this pager has focus
            layout = get_app().layout
            if layout.has_focus(self):
                previous_control = layout.previous_control
                if previous_control in layout.find_all_controls():
                    layout.focus(previous_control)

    @property
    def state(self) -> "Optional[PagerState]":
        """Return the pager's current state."""
        return self._state

    @state.setter
    def state(self, new: "Optional[PagerState]") -> "None":
        """Set the pager's current state."""
        self._state = new
        if new is not None:
            self.output.json = new.response
            self.output.update()

    def __pt_container__(self):
        """Return the pager container."""
        return self.container

    # ################################## Commands #####################################

    @staticmethod
    @add_cmd(filter=pager_has_focus)
    def _close_pager() -> "None":
        """Close the pager."""
        app = get_app()
        if app.pager is not None:
            app.pager.hide()

    # ################################# Keybindings ###################################

    register_bindings(
        {
            "euporie.core.widgets.pager.Pager": {
                "close-pager": ["escape", "q"],
            }
        }
    )



"""Defines a status-bar widget."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import FormattedTextControl

from euporie.core.app import get_app
from euporie.core.config import add_setting
from euporie.core.filters import is_searching

if TYPE_CHECKING:
    from prompt_toolkit.filters.base import FilterOrBool
    from prompt_toolkit.layout.containers import AnyContainer


class StatusBar:
    """A status bar which shows the status of the current tab."""

    def __init__(self, extra_filter: "FilterOrBool" = True) -> "None":
        """Create a new status bar instance."""
        self.app = get_app()
        self.container = ConditionalContainer(
            content=VSplit(
                [
                    Window(
                        FormattedTextControl(
                            lambda: self.app.format_status(part="left")
                        ),
                        style="class:status",
                    ),
                    Window(
                        FormattedTextControl(
                            lambda: self.app.format_status(part="right")
                        ),
                        style="class:status.right",
                        align=WindowAlign.RIGHT,
                    ),
                ],
                height=1,
            ),
            filter=get_app().config.filter("show_status_bar")
            & ~is_searching
            & to_filter(extra_filter),
        )

    def __pt_container__(self) -> "AnyContainer":
        """Return the widget's container."""
        return self.container

    # ################################### Settings ####################################

    add_setting(
        name="show_status_bar",
        flags=["--show-status-bar"],
        type_=bool,
        title="status bar",
        help_="Show the status bar",
        default=True,
        schema={
            "type": "boolean",
        },
        description="""
            Whether the status bar should be shown at the bottom of the screen.
        """,
    )



"""Defines a cell object with input are and rich outputs, and related objects."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters import (
    Condition,
    buffer_has_focus,
    has_focus,
    has_selection,
    is_done,
    is_searching,
    to_filter,
)
from prompt_toolkit.key_binding.key_bindings import merge_key_bindings
from prompt_toolkit.layout.containers import ConditionalContainer
from prompt_toolkit.layout.margins import ConditionalMargin
from prompt_toolkit.layout.processors import (  # HighlightSearchProcessor,
    ConditionalProcessor,
    DisplayMultipleCursors,
    HighlightIncrementalSearchProcessor,
    HighlightMatchingBracketProcessor,
    HighlightSelectionProcessor,
    TabsProcessor,
)
from prompt_toolkit.lexers import DynamicLexer, PygmentsLexer
from prompt_toolkit.widgets import TextArea
from pygments.lexers import get_lexer_by_name

from euporie.core.app import get_app
from euporie.core.commands import add_cmd
from euporie.core.config import add_setting
from euporie.core.filters import buffer_is_code
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.margins import NumberedDiffMargin, OverflowMargin, ScrollbarMargin
from euporie.core.processors import (
    AppendLineAutoSuggestion,
    ShowTrailingWhiteSpaceProcessor,
)
from euporie.core.suggest import ConditionalAutoSuggestAsync
from euporie.core.widgets.pager import PagerState

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, Sequence, Union

    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.key_binding.key_bindings import KeyBindingsBase
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.layout import FocusableElement
    from prompt_toolkit.layout.margins import Margin

    from euporie.core.tabs.base import KernelTab


log = logging.getLogger(__name__)


class KernelInput(TextArea):
    """Kernel input text areas.

    A customized text area for the cell input.
    """

    def __init__(
        self,
        kernel_tab: "KernelTab",
        *args: "Any",
        left_margins: "Optional[Sequence[Margin]]" = None,
        right_margins: "Optional[Sequence[Margin]]" = None,
        on_text_changed: "Optional[Callable[[Buffer], None]]" = None,
        on_cursor_position_changed: "Optional[Callable[[Buffer], None]]" = None,
        tempfile_suffix: "Union[str, Callable[[], str]]" = "",
        key_bindings: "Optional[KeyBindingsBase]" = None,
        enable_history_search: "Optional[FilterOrBool]" = False,
        focusable: "FilterOrBool" = True,
        wrap_lines: "FilterOrBool" = False,
        complete_while_typing: "FilterOrBool" = True,
        autosuggest_while_typing: "FilterOrBool" = True,
        validate_while_typing: "Optional[FilterOrBool]" = None,
        **kwargs: "Any",
    ) -> "None":
        """Initiate the cell input box."""
        self.kernel_tab = kernel_tab
        app = kernel_tab.app

        kwargs.setdefault("wrap_lines", wrap_lines)
        kwargs.setdefault("style", "class:kernel-input")
        kwargs.setdefault("history", kernel_tab.history)
        kwargs.setdefault("search_field", app.search_bar)
        kwargs.setdefault("focus_on_click", True)
        kwargs.setdefault(
            "input_processors",
            [
                ConditionalProcessor(
                    HighlightIncrementalSearchProcessor(),
                    filter=is_searching,
                ),
                # HighlightSearchProcessor(),
                HighlightSelectionProcessor(),
                DisplayMultipleCursors(),
                HighlightMatchingBracketProcessor(),
                TabsProcessor(char1="⇥", char2="┈"),
                ShowTrailingWhiteSpaceProcessor(),
            ],
        )
        kwargs.setdefault("completer", kernel_tab.completer)
        kwargs.setdefault(
            "lexer",
            DynamicLexer(
                lambda: PygmentsLexer(
                    get_lexer_by_name(self.kernel_tab.language).__class__,
                    sync_from_start=False,
                )
            ),
        )
        kwargs.setdefault(
            "auto_suggest",
            ConditionalAutoSuggestAsync(
                self.kernel_tab.suggester,
                filter=to_filter(autosuggest_while_typing)
                & Condition(lambda: app.config.autosuggest),
            ),
        )
        kwargs["complete_while_typing"] = Condition(
            lambda: app.config.autocomplete
        ) & to_filter(complete_while_typing)

        super().__init__(*args, **kwargs)

        if validate_while_typing:
            self.buffer.validate_while_typing = to_filter(validate_while_typing)

        self.control.include_default_input_processors = False
        if on_text_changed:
            self.buffer.on_text_changed += on_text_changed
        if on_cursor_position_changed:
            self.buffer.on_cursor_position_changed += on_cursor_position_changed
        self.buffer.tempfile_suffix = tempfile_suffix or kernel_tab.kernel_lang_file_ext

        if enable_history_search is not None:
            self.buffer.enable_history_search = to_filter(enable_history_search)

        self.has_focus = has_focus(self)

        # Replace the autosuggest processor
        # Skip type checking as PT should use "("Optional[Sequence[Processor]]"
        # instead of "Optional[list[Processor]]"
        # TODO make a PR for this
        self.control.input_processors[0] = ConditionalProcessor(  # type: ignore
            AppendLineAutoSuggestion(),
            has_focus(self.buffer) & ~is_done,
        )

        # Set style
        style = kwargs.get("style", "")
        self.window.style = lambda: (
            "class:text-area " + ("class:focused " if self.has_focus() else "") + style
        )

        # Add configurable line numbers
        self.window.left_margins = left_margins or [
            ConditionalMargin(
                NumberedDiffMargin(),
                Condition(lambda: self.kernel_tab.app.config.line_numbers),
            )
        ]
        self.window.right_margins = right_margins or [
            OverflowMargin(),
            ScrollbarMargin(),
        ]

        self.window.cursorline = self.has_focus

        # Set extra key-bindings
        widgets_key_bindings = load_registered_bindings(
            "euporie.core.widgets.inputs.KernelInput"
        )
        if key_bindings:
            widgets_key_bindings = merge_key_bindings(
                [key_bindings, widgets_key_bindings]
            )
        self.control.key_bindings = widgets_key_bindings

    def inspect(self) -> "None":
        """Get contextual help for the current cursor position in the current cell."""
        code = self.buffer.text
        cursor_pos = self.buffer.cursor_position

        pager = self.kernel_tab.app.pager
        assert pager is not None

        if pager.visible() and pager.state is not None:
            if pager.state.code == code and pager.state.cursor_pos == cursor_pos:
                pager.focus()
                return

        def _cb(response: "dict") -> "None":
            assert pager is not None
            prev_state = pager.state
            new_state = PagerState(
                code=code,
                cursor_pos=cursor_pos,
                response=response,
            )
            if prev_state != new_state:
                pager.state = new_state
                self.kernel_tab.app.invalidate()

        self.kernel_tab.kernel.inspect(
            code=code,
            cursor_pos=cursor_pos,
            callback=_cb,
        )

    # ################################### Settings ####################################

    add_setting(
        name="line_numbers",
        flags=["--line-numbers"],
        type_=bool,
        help_="Show or hide line numbers",
        default=True,
        description="""
            Whether line numbers are shown by default.
        """,
        hooks=[lambda x: get_app().refresh()],
    )

    add_setting(
        name="autocomplete",
        flags=["--autocomplete"],
        type_=bool,
        help_="Provide completions suggestions automatically",
        default=False,
        description="""
            Whether to automatically suggestion completions while typing in code cells.
        """,
    )

    add_setting(
        name="autosuggest",
        flags=["--autosuggest"],
        type_=bool,
        help_="Provide line completion suggestions",
        default=True,
        description="""
            Whether to automatically suggestion line content while typing in code cells.
        """,
    )

    add_setting(
        name="autoinspect",
        flags=["--autoinspect"],
        type_=bool,
        help_="Display contextual help automatically",
        default=False,
        description="""
            Whether to automatically display contextual help when navigating through code cells.
        """,
    )

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus & ~has_selection,
    )
    def _show_contextual_help() -> "None":
        """Displays contextual help."""
        from euporie.core.tabs.notebook import BaseNotebook

        nb = get_app().tab
        if isinstance(nb, BaseNotebook):
            nb.cell.input_box.inspect()

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.widgets.inputs.KernelInput": {
                "show-contextual-help": "s-tab",
            }
        }
    )


class StdInput:
    """A widget to accept kernel input."""

    def __init__(self, kernel_tab: "KernelTab") -> "None":
        """Create a new kernel input box."""
        from euporie.core.widgets.forms import LabelledWidget, Text

        self.kernel_tab = kernel_tab
        self.last_focused: "Optional[FocusableElement]" = None

        self.password = False
        self.prompt: "Optional[str]" = None
        self.active = False
        self.visible = Condition(lambda: self.active)

        text = Text(
            multiline=False,
            accept_handler=self.accept,
            password=Condition(lambda: self.password),
            style="class:input",
        )
        self.window = text.text_area.window
        self.container = ConditionalContainer(
            LabelledWidget(
                body=text,
                label=lambda: self.prompt or ">>>",
            ),
            filter=self.visible,
        )

    def accept(self, buffer: "Buffer") -> "bool":
        """Send the input to the kernel and hide the input box."""
        if self.kernel_tab.kernel.kc is not None:
            self.kernel_tab.kernel.kc.input(buffer.text)
        # Cleanup
        self.active = False

        buffer.text = ""
        if self.last_focused:
            try:
                get_app().layout.focus(self.last_focused)
            except ValueError:
                pass
        return True

    def get_input(
        self,
        prompt: "str" = "Please enter a value: ",
        password: "bool" = False,
    ) -> "None":
        """Prompts the user for input and sends the result to the kernel."""
        self.password = password
        self.prompt = prompt

        # Set this first so the height of the cell includes the input box if it gets
        # rendered when we scroll to it
        self.active = True
        # Remember what was focused before
        app = get_app()
        layout = app.layout
        self.last_focused = layout.current_control

        # Try focusing the input box - we create an asynchronous task which will
        # probably run after the next render, when the stdin_box is recognised as being
        # in the layout. This doesn't always work (depending on timing), does usually.
        try:
            layout.focus(self)
        finally:

            async def _focus_input() -> "None":
                # Focus the input box
                if self.window in layout.visible_windows:
                    layout.focus(self)
                # Redraw the screen to show it as focused
                app.invalidate()

            app.create_background_task(_focus_input())

    def __pt_container__(self) -> "AnyContainer":
        """Return the input's container."""
        return self.container



"""Allows access to the current running application."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app as ptk_get_app

if TYPE_CHECKING:
    from euporie.core.app import BaseApp


def get_app() -> "BaseApp":
    """Get the current application."""
    app = ptk_get_app()
    return cast("BaseApp", app)



"""Contains containers which display children at full height vertially stacked."""

from __future__ import annotations

import asyncio
import contextvars
import logging
import weakref
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import is_searching
from prompt_toolkit.layout.containers import (
    Container,
    ScrollOffsets,
    Window,
    WindowRenderInfo,
    to_container,
)
from prompt_toolkit.layout.controls import UIContent
from prompt_toolkit.layout.dimension import Dimension, to_dimension
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.layout.screen import Char, Screen, WritePosition
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType

if TYPE_CHECKING:
    from typing import Callable, Optional, Sequence, Union

    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension

    from euporie.core.widgets.cell import Cell

    MouseHandler = Callable[[MouseEvent], object]

log = logging.getLogger(__name__)


class ChildRenderInfo:
    """A class which holds information about a :py:class:`ScrollingContainer` child."""

    def __init__(self, parent: "ScrollingContainer", child: "Cell") -> "None":
        """Initiates the :py:class:`ChildRenderInfo` object.

        Args:
            parent: The parent scrolling container this relates to
            child: The child container (notebook cell) this information concerns

        """
        self.parent = parent
        self.child = weakref.proxy(child)
        child.meta = self
        self.container = to_container(child)

        self.screen = Screen(default_char=Char(char=" "))
        self.mouse_handlers = MouseHandlers()

        self.height = 0
        self.width = 0

        self.refresh = True

        self.win_render_infos: "dict[Window, WindowRenderInfo]" = {}

    def render(
        self,
        available_width: "int",
        available_height: "int",
        style: "str" = "",
    ) -> "None":
        """Renders the child container at a given size.

        Args:
            available_width: The height available for rendering
            available_height: The width available for rendering
            style: The parent style to apply when rendering

        """
        if self.refresh:
            self.refresh = False
            # log.debug("Re-rendering cell %s", self.child.index)
            self.height = self.container.preferred_height(
                available_width, available_height
            ).preferred
            # TODO - allow horizontal scrolling too
            # self.width = self.container.preferred_width(available_width).width
            self.width = available_width

            self.container.write_to_screen(
                self.screen,
                self.mouse_handlers,
                WritePosition(
                    0,
                    0,
                    self.width,
                    self.height,
                ),
                style,
                erase_bg=True,
                z_index=0,
            )
            self.screen.draw_all_floats()

    def blit(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        left: "int",
        top: "int",
        cols: "slice",
        rows: "slice",
    ) -> "None":
        """Copies the rendered child from the local screen to the main screen.

        All locations are adjusted, allowing the pre-rendered child to be placed at any
        location on the main screen.

        Args:
            screen: The main screen to copy the pre-rendered screen data to
            mouse_handlers: The mouse handler collection to copy the pre-rendered
                handelers to
            left: The left-most column in which to start placing the data
            top: The upper row in which to start placing the data
            cols: The columns to copy
            rows: The rows to copy

        """
        # Copy write positions
        for win, wp in self.screen.visible_windows_to_write_positions.items():
            new_wp = WritePosition(
                xpos=wp.xpos + left,
                ypos=wp.ypos + top,
                width=wp.width,
                height=max(
                    0,
                    min(
                        wp.height + wp.ypos - max(wp.ypos, rows.start),
                        min(wp.height + wp.ypos, rows.stop) - wp.ypos,
                    ),
                ),
            )
            screen.visible_windows_to_write_positions[win] = new_wp

            # Modify render info
            if win.render_info is not None:
                info = self.win_render_infos.setdefault(win, win.render_info)
                # info.visible_line_to_row_col = {
                # line: (y + info._y_offset, new_wp.xpos + info._x_offset)
                # for line, y in enumerate(
                # range(new_wp.ypos, new_wp.ypos + new_wp.height)
                # )
                # }
                win.render_info = WindowRenderInfo(
                    window=win,
                    ui_content=info.ui_content,
                    horizontal_scroll=0,
                    vertical_scroll=info.vertical_scroll,
                    window_width=info.window_width,
                    window_height=info.window_height,
                    configured_scroll_offsets=info.configured_scroll_offsets,
                    visible_line_to_row_col=info.visible_line_to_row_col,
                    rowcol_to_yx={
                        (row, col): (y + top, x + left)
                        for (row, col), (y, x) in info._rowcol_to_yx.items()
                    },
                    x_offset=info._x_offset + left,
                    y_offset=info._y_offset + top,
                    wrap_lines=info.wrap_lines,
                )

        mouse_handler_wrappers: "dict[MouseHandler, MouseHandler]" = {}

        def _wrap_mouse_handler(
            handler: "Callable",
        ) -> "MouseHandler":
            if mouse_handler := mouse_handler_wrappers.get(handler):
                return mouse_handler
            else:

                def wrapped_mouse_handler(
                    mouse_event: "MouseEvent",
                ) -> "NotImplementedOrNone":
                    response: "NotImplementedOrNone" = NotImplemented
                    if handler:
                        new_event = MouseEvent(
                            position=Point(
                                x=mouse_event.position.x - left,
                                y=mouse_event.position.y - top,
                            ),
                            event_type=mouse_event.event_type,
                            button=mouse_event.button,
                            modifiers=mouse_event.modifiers,
                        )
                        response = handler(new_event)
                    # This would work if windows returned NotImplemented when scrolled
                    # to the start or end
                    # if response is NotImplemented:
                    if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
                        response = self.parent.scroll(-1)
                    elif mouse_event.event_type == MouseEventType.SCROLL_UP:
                        response = self.parent.scroll(1)
                    return response

                mouse_handler_wrappers[handler] = wrapped_mouse_handler
                return wrapped_mouse_handler

        # Copy screen contents
        input_db = self.screen.data_buffer
        input_zwes = self.screen.zero_width_escapes
        input_mhs = self.mouse_handlers.mouse_handlers

        output_dbs = screen.data_buffer
        output_zwes = screen.zero_width_escapes
        output_mhs = mouse_handlers.mouse_handlers
        for y in range(max(0, rows.start), min(rows.stop, self.height)):
            output_dbs_row = output_dbs[top + y]
            output_zwes_row = output_zwes[top + y]
            output_mhs_row = output_mhs[top + y]

            input_db_row = input_db[y]
            input_zwes_row = input_zwes[y]
            input_mhs_row = input_mhs[y]
            for x in range(max(0, cols.start), min(cols.stop, self.width)):
                # Data
                output_dbs_row[left + x] = input_db_row[x]
                # Escape sequences
                output_zwes_row[left + x] = input_zwes_row[x]
                # Mouse handlers
                output_mhs_row[left + x] = _wrap_mouse_handler(input_mhs_row[x])

        # Copy cursors
        if self.screen.show_cursor:
            for window, point in self.screen.cursor_positions.items():
                if get_app().layout.current_control == window.content:
                    assert window.render_info is not None
                    if (
                        window.render_info.ui_content.show_cursor
                        and not window.always_hide_cursor()
                    ):
                        if point.x in range(cols.start, cols.stop) and point.y in range(
                            rows.start, rows.stop
                        ):
                            screen.cursor_positions[window] = Point(
                                x=left + point.x, y=top + point.y
                            )
                            screen.show_cursor = True
        # Copy menu positions
        for window, point in self.screen.menu_positions.items():
            screen.menu_positions[window] = Point(x=left + point.x, y=top + point.y)


class ScrollingContainer(Container):
    """A scrollable container which renders only the currently visible children."""

    def __init__(
        self,
        children_func: "Callable",
        height: "AnyDimension" = None,
        width: "AnyDimension" = None,
        style: "Union[str, Callable[[], str]]" = "",
    ):
        """Initiates the `ScrollingContainer`."""
        self.children_func = children_func
        self._children: "list[Cell]" = []
        self.refresh_children = True
        self.pre_rendered = 0.0

        self._selected_slice = slice(
            0, 1
        )  # The index of the currently selected children
        self._selected_child_render_infos: "list[ChildRenderInfo]"
        self.selected_child_position: "int" = 0

        self.child_render_infos: "dict[int, ChildRenderInfo]" = (
            {}
        )  # Holds child container wrappers
        self.visible_indicies: "set[int]" = {0}
        self.index_positions: "dict[int, Optional[int]]" = {}

        self.last_write_position = WritePosition(0, 0, 0, 0)

        self.height = to_dimension(height).preferred
        self.width = to_dimension(width).preferred
        self.style = style

        self.scroll_to_cursor = False
        self.scrolling = 0

    def pre_render_children(self, width: "int", height: "int") -> "None":
        """Render all unrendered children in a background thread."""
        # Copy the current context so ``get_app()`` works in the thread
        ctx = contextvars.copy_context()

        def render_in_thread() -> "None":
            """Create a new event loop in the thread."""
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def render_run_in_loop() -> "None":
                """Render all children sequentially."""
                n_children = len(self.children)
                for i in range(n_children):
                    self.get_child_render_info(i).render(width, height)
                    self.pre_rendered = i / n_children
                    get_app().invalidate()
                self.pre_rendered = 1.0
                get_app().invalidate()

            loop.run_until_complete(render_run_in_loop())

        async def trigger_render() -> "None":
            """Use an executor thread from the current event loop."""
            await asyncio.get_event_loop().run_in_executor(
                None, ctx.run, render_in_thread
            )

        get_app().create_background_task(trigger_render())

    def reset(self) -> "None":
        """Reset the state of this container and all the children."""
        meta_data = list(self.child_render_infos.values())
        for meta in meta_data:
            meta.container.reset()
            meta.refresh = True

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Do not provide a preferred width - grow to fill the available space."""
        if self.width is not None:
            return Dimension(min=1, preferred=self.width, weight=1)
        return Dimension()

    def preferred_height(self, width: int, max_available_height: int) -> "Dimension":
        """Return the preferred height only if one is provided."""
        if self.height is not None:
            return Dimension(min=1, preferred=self.height)
        return Dimension()

    @property
    def children(
        self,
    ) -> "Sequence[Cell]":  # Sequence[Union[Container, MagicContainer]]":
        """Return the current children of this container instance."""
        if self.refresh_children:
            self._children = self.children_func()
            self.refresh_children = False
            # Clean up metacache
            for child_hash in set(self.child_render_infos) - set(
                map(hash, self._children)
            ):
                del self.child_render_infos[child_hash]
            # Clean up positions
            self.index_positions = {
                i: pos
                for i, pos in self.index_positions.items()
                if i < len(self._children)
            }
        return self._children

    def _set_selected_slice(
        self, new_slice: "slice", force: "bool" = False, scroll: "bool" = True
    ) -> None:
        # Only update the selected child if it was not selected before
        if force or new_slice != self._selected_slice:
            self.refresh_children = True
            # Ensure new selected slice is valid
            new_slice = self.validate_slice(new_slice)
            # Scroll into view
            if scroll:
                self.scroll_to(new_slice.start)
            # Request a refresh of the previously selected children
            for render_info in self._selected_child_render_infos:
                if render_info:
                    render_info.refresh = True
            # Get the first selected child and focus it
            child = self.children[new_slice.start]
            app = get_app()
            if not app.layout.has_focus(child):
                app.layout.focus(child)
            # Track which child was selected
            self._selected_slice = new_slice

    @property
    def selected_slice(self) -> "slice":
        """Returns the currently selected slice."""
        return self._selected_slice

    @selected_slice.setter
    def selected_slice(self, new_slice: "slice") -> "None":
        """Sets the currently selected child index.

        Args:
            new_slice: The slice of the children to select.
        """
        self._set_selected_slice(new_slice)

    def validate_slice(self, slice_: "slice") -> "slice":
        """Ensures a slice describes a valid range of children."""
        start = min(max(slice_.start, 0), len(self.children) - 1)
        stop = slice_.stop
        if stop == -1:
            stop = None
        if stop is not None:
            stop = min(max(slice_.stop, int(start == 0)), len(self.children))
        return slice(start, stop, slice_.step)

    @property
    def selected_indices(self) -> "list[int]":
        """Returns in indices of the currently selected children."""
        return list(range(*self.selected_slice.indices(len(self._children))))

    def get_child_render_info(self, index: "Optional[int]" = None) -> "ChildRenderInfo":
        """Return a rendered instance of the child at the given index.

        If no index is given, the currently selected child is returned.

        Args:
            index: The index of the child to return.

        Returns:
            A rendered instance of the child.

        """
        if index is None:
            index = self._selected_slice.start
        child = self.children[index]
        child_hash = hash(child)
        if child_hash not in self.child_render_infos:
            child_render_info = ChildRenderInfo(self, child)
            self.child_render_infos[child_hash] = child_render_info
        else:
            child_render_info = self.child_render_infos[child_hash]
        return child_render_info

    def scroll(self, n: "int") -> "NotImplementedOrNone":
        """Scrolls up or down a number of rows.

        Args:
            n: The number of rows to scroll, negative for up, positive for down

        Returns:
            :py:const:`NotImplemented` is scrolling is not allowed, otherwise
                :py:const:`None`

        """
        # self.refresh_children = True
        if n > 0:
            if (
                min(self.visible_indicies) == 0
                and self.index_positions[0] is not None
                and self.index_positions[0] + self.scrolling + n > 0
            ):
                return NotImplemented
        elif n < 0:
            bottom_index = len(self.children) - 1
            if bottom_index in self.visible_indicies:
                bottom_child = self.get_child_render_info(bottom_index)
                bottom_pos = self.index_positions[bottom_index]
                if (
                    bottom_pos is not None
                    and bottom_pos + bottom_child.height + self.scrolling + n
                    < self.last_write_position.height
                ):
                    return NotImplemented

        # Very basic scrolling acceleration
        self.scrolling += n
        return None

    def mouse_scroll_handler(self, mouse_event: "MouseEvent") -> "NotImplementedOrNone":
        """A mouse handler to scroll the pane."""
        if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            return self.scroll(-1)
        elif mouse_event.event_type == MouseEventType.SCROLL_UP:
            return self.scroll(1)
        else:
            return NotImplemented
        return None

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: str,
        erase_bg: bool,
        z_index: Optional[int],
    ) -> "None":
        """Write the actual content to the screen.

        Children are rendered only if they are visible and have changed, and the output
        is cached to a separate screen object stored in a :py::class:`ChildRenderInfo`. The
        cached rendering of the children which are actually visible are then copied to
        the screen. This results in faster rendering of the scrolling container, and
        makes scrolling more performant.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.
        """
        ypos = write_position.ypos
        xpos = write_position.xpos
        available_width = write_position.width
        available_height = write_position.height

        if not self.pre_rendered:
            self.pre_render_children(available_width, available_height)

        # Record children which are currently visible
        visible_indicies = set()

        # Force the selected children to refresh
        selected_indices = self.selected_indices
        self._selected_child_render_infos = []
        for index in selected_indices:
            render_info = self.get_child_render_info(index)
            # Do not bother to re-render children if we are scrolling
            if not self.scrolling:
                render_info.refresh = True
            self._selected_child_render_infos.append(render_info)
            self.index_positions[index] = None

        # Refresh all children if the width has changed
        if (
            self.last_write_position is not None
            and self.last_write_position.width != write_position.width
        ):
            for child_render_info in self.child_render_infos.values():
                child_render_info.refresh = True

        # Refresh visible children if searching
        if is_searching():
            for index in self.visible_indicies:
                self.get_child_render_info(index).refresh = True

        # Scroll to make the cursor visible
        if self.scroll_to_cursor:
            selected_child_render_info = self._selected_child_render_infos[0]
            selected_child_render_info.render(
                available_width=available_width,
                available_height=available_height,
                style=f"{parent_style} {self.style}",
            )
            current_window = get_app().layout.current_window
            cursor_position = selected_child_render_info.screen.cursor_positions.get(
                current_window
            )
            if cursor_position:
                cursor_row = self.selected_child_position + cursor_position.y
                if cursor_row < 0:
                    self.selected_child_position -= cursor_row
                elif available_height <= cursor_row:
                    self.selected_child_position -= cursor_row - available_height + 1
            # Set this to false again, allowing us to scroll the cursor out of view
            self.scroll_to_cursor = False

        # Adjust scrolling offset
        if self.scrolling:
            heights = [
                # Ensure unrendered cells have at least some height
                self.get_child_render_info(index).height or 1
                for index in range(len(self._children))
            ]
            heights_above = sum(heights[: self._selected_slice.start])
            new_child_position = self.selected_child_position + self.scrolling
            # Do not allow scrolling if there is no overflow
            if sum(heights) < available_height:
                self.selected_child_position = heights_above
            else:
                # Prevent overscrolling at the top of the document
                overscroll = heights_above - new_child_position
                if overscroll < 0:
                    # Scroll as far as we are able without overscrolling
                    self.scrolling = max(0, self.scrolling + overscroll)
                # Prevent underscrolling at the bottom
                elif overscroll > 0:
                    heights_below = sum(heights[self._selected_slice.start :])
                    underscroll = new_child_position + heights_below - available_height
                    if underscroll < 0:
                        # Scroll as far as we are able without underscrolling
                        self.scrolling = min(0, self.scrolling - underscroll)
            self.selected_child_position += self.scrolling

        # Blit first selected child and those below it that are on screen
        line = self.selected_child_position
        for i in range(self._selected_slice.start, len(self.children)):
            child_render_info = self.get_child_render_info(i)
            child_render_info.render(
                available_width=available_width,
                available_height=available_height,
                style=f"{parent_style} {self.style}",
            )
            if 0 < line + child_render_info.height and line < available_height:
                self.index_positions[i] = line
                child_render_info.blit(
                    screen=screen,
                    mouse_handlers=mouse_handlers,
                    left=xpos,
                    top=ypos + line,
                    cols=slice(0, available_width),
                    rows=slice(
                        max(0, 0 - line),
                        min(child_render_info.height, available_height - line),
                    ),
                )
                visible_indicies.add(i)
            line += child_render_info.height
            if line >= available_height:
                break
        else:
            if line < available_height:
                Window(char=" ").write_to_screen(
                    screen,
                    mouse_handlers,
                    WritePosition(
                        xpos, ypos + line, available_width, available_height - line
                    ),
                    parent_style,
                    erase_bg,
                    z_index,
                )
                for y in range(ypos + line, ypos + available_height):
                    for x in range(xpos, xpos + available_width):
                        mouse_handlers.mouse_handlers[y][x] = self.mouse_scroll_handler
        # Blit children above the selected that are on screen
        line = self.selected_child_position
        for i in range(self._selected_slice.start - 1, -1, -1):
            child_render_info = self.get_child_render_info(i)
            child_render_info.render(
                available_width=available_width,
                available_height=available_height,
                style=f"{parent_style} {self.style}",
            )
            line -= child_render_info.height
            if 0 < line + child_render_info.height and line < available_height:
                self.index_positions[i] = line
                child_render_info.blit(
                    screen=screen,
                    mouse_handlers=mouse_handlers,
                    left=xpos,
                    top=ypos + line,
                    cols=slice(0, available_width),
                    rows=slice(
                        max(0, 0 - line),
                        min(child_render_info.height, available_height - line),
                    ),
                )
                visible_indicies.add(i)
            if line <= 0:
                break
        else:
            if line > 0:
                Window(char=" ").write_to_screen(
                    screen,
                    mouse_handlers,
                    WritePosition(xpos, ypos, available_width, line),
                    parent_style,
                    erase_bg,
                    z_index,
                )
                for y in range(ypos, ypos + line):
                    for x in range(xpos, xpos + available_width):
                        mouse_handlers.mouse_handlers[y][x] = self.mouse_scroll_handler

        # Dont bother drawing floats
        # screen.draw_all_floats()

        # Ensure the focused child is always in the layout
        visible_indicies.add(self._selected_slice.start)

        # Update which children will appear in the layout
        self.visible_indicies = visible_indicies
        # Record where the contain was last drawn so we can determine if cell outputs
        # are partially obscured
        self.last_write_position = write_position

        # Calculate scrollbar info
        sizes = self.known_sizes
        avg_size = sum(sizes.values()) / len(sizes)
        n_children = len(self.children)
        for i in range(n_children):
            if i not in sizes:
                sizes[i] = int(avg_size)
        content_height = max(sum(sizes.values()), 1)

        # Mock up a WindowRenderInfo so we can draw a scrollbar margin
        self.render_info = WindowRenderInfo(
            window=cast("Window", self),
            ui_content=UIContent(line_count=content_height),
            horizontal_scroll=0,
            vertical_scroll=self.vertical_scroll,
            window_width=available_width,
            window_height=available_height,
            configured_scroll_offsets=ScrollOffsets(),
            visible_line_to_row_col={i: (i, 0) for i in range(available_height)},
            rowcol_to_yx={},
            x_offset=xpos,
            y_offset=ypos,
            wrap_lines=False,
        )
        # Signal that we are no longer scrolling
        self.scrolling = 0

    @property
    def vertical_scroll(self) -> "int":
        """The best guess at the absolute vertical scroll position."""
        return (
            sum(list(self.known_sizes.values())[: self._selected_slice.start])
            - self.selected_child_position
        )

    @vertical_scroll.setter
    def vertical_scroll(self, value: "int") -> "None":
        """Set the absolute vertical scroll position."""
        self.selected_child_position = (
            sum(list(self.known_sizes.values())[: self._selected_slice.start]) - value
        )

    def get_children(self) -> "list[Container]":
        """Return the list of currently visible children to include in the layout."""
        return [
            self.get_child_render_info(i).container
            for i in self.visible_indicies
            if i < len(self.children)
        ]

    def get_child(self, index: "Optional[int]" = None) -> "AnyContainer":
        """Return a rendered instance of the child at the given index.

        If no index is given, the currently selected child is returned.

        Args:
            index: The index of the child to return.

        Returns:
            A rendered instance of the child.

        """
        if index is None:
            index = self.selected_slice.start
        return self.children[index]

    def scroll_to(self, index: "int") -> "None":
        """Scroll a child into view.

        Args:
            index: The child index to scroll into view

        """
        child_render_info = self.get_child_render_info(index)
        child_render_info.render(
            self.last_write_position.width, self.last_write_position.height
        )

        if index in self.visible_indicies:
            new_top = self.index_positions[index]
        else:
            if index < self._selected_slice.start:
                # new_top = 0
                new_top = max(
                    0, child_render_info.height - self.last_write_position.height
                )
            elif index == self._selected_slice.start:
                new_top = self.selected_child_position
            elif index > self._selected_slice.start:
                last_index = max(
                    k for k, v in self.index_positions.items() if v is not None
                )
                new_top = max(
                    min(
                        self.last_write_position.height,
                        (self.index_positions[last_index] or 0)
                        + self.get_child_render_info(last_index).height,
                    ),
                    0,
                )

        if new_top is None or new_top < 0:
            self.selected_child_position = min(
                0, self.last_write_position.height - child_render_info.height
            )
        elif new_top > self.last_write_position.height - child_render_info.height:
            self.selected_child_position = max(
                0,
                self.last_write_position.height - child_render_info.height,
            )
        else:
            self.selected_child_position = new_top

    @property
    def known_sizes(self) -> "dict[int, int]":
        """A dictionary mapping child indices to height values."""
        sizes = {}
        for i, child in enumerate(self.children):
            child_hash = hash(child)
            if child_render_info := self.child_render_infos.get(child_hash):
                if child_render_info.height:
                    sizes[i] = child_render_info.height
        return sizes

    def _scroll_up(self) -> "None":
        """Scroll up one line: for compatibility with :py:class:`Window`."""
        self.scroll(1)

    def _scroll_down(self) -> "None":
        """Scroll down one line: for compatibility with :py:class:`Window`."""
        self.scroll(-1)


class PrintingContainer(Container):
    """A container which displays all it's children in a vertical list."""

    def __init__(
        self,
        children: "Union[Callable, Sequence[AnyContainer]]",
        width: "AnyDimension" = None,
        key_bindings: "Optional[KeyBindingsBase]" = None,
    ):
        """Initiate the container."""
        self.width = width
        self.rendered = False
        self._children = children
        self.key_bindings = key_bindings

    def get_key_bindings(self) -> Optional[KeyBindingsBase]:
        """Return the container's key bindings."""
        return self.key_bindings

    @property
    def children(self) -> "Sequence[AnyContainer]":
        """Returns the container's children."""
        if callable(self._children):
            children = self._children()
        else:
            children = self._children
        return children or [Window()]

    def get_children(self) -> "list[Container]":
        """Returns a list of all child containers."""
        return list(map(to_container, self.children))

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: "str",
        erase_bg: "bool",
        z_index: "Optional[int]",
    ) -> "None":
        """Render the container to a `Screen` instance.

        All children are rendered vertically in sequence.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        xpos = write_position.xpos
        ypos = write_position.ypos

        children = self.get_children()
        for child in children:
            height = child.preferred_height(write_position.width, 999999).preferred
            child.write_to_screen(
                screen,
                mouse_handlers,
                WritePosition(xpos, ypos, write_position.width, height),
                parent_style,
                erase_bg,
                z_index,
            )
            ypos += height

    def preferred_height(self, width: int, max_available_height: int) -> "Dimension":
        """Returns the preferred height, equal to the sum of the child heights."""
        return Dimension(
            min=1,
            preferred=sum(
                [
                    c.preferred_height(width, max_available_height).preferred
                    for c in self.get_children()
                ]
            ),
        )

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Calculates and returns the desired width for this container."""
        if self.width is not None:
            dim = to_dimension(self.width).preferred
            return Dimension(max=dim, preferred=dim)
        else:
            return Dimension(max_available_width)

    def reset(self) -> "None":
        """Reset the state of this container and all the children.

        Does nothing as this container is used for dumping output.
        """
        pass



"""Utility functions for format converters."""

from __future__ import annotations

import io
import logging
import os
import subprocess  # noqa S404 - Security implications have been considered
import tempfile
from functools import partial, reduce
from importlib import import_module
from shutil import which
from typing import TYPE_CHECKING

import imagesize
from prompt_toolkit.filters import Condition, to_filter

from euporie.core.app import get_app
from euporie.core.convert.base import convert

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from prompt_toolkit.filters import Filter

log = logging.getLogger(__name__)


def commands_exist(*cmds: "str") -> "Filter":
    """Verifies a list of external commands exist on the system."""
    filters = [
        Condition(partial(lambda x: bool(which(cmd)), cmd))  # noqa: B023
        for cmd in cmds
    ]
    return reduce(lambda a, b: a & b, filters, to_filter(True))


def have_modules(*modules: "str") -> "Filter":
    """Verifies a list of python modules are importable."""

    def try_import(module: "str") -> "bool":
        try:
            import_module(module)
        except ModuleNotFoundError:
            return False
        else:
            return True

    filters = [Condition(partial(try_import, module)) for module in modules]
    return reduce(lambda a, b: a & b, filters, to_filter(True))


def call_subproc(
    data: "Union[str, bytes]", cmd: "list[Any]", use_tempfile: "bool" = False
) -> "bytes":
    """Call the command as a subprocess and return it's output as bytes.

    Args:
        data: The data to pass to the subprocess
        cmd: The command and arguments to call
        use_tempfile: If True, the command saves its output to a file, not stdout

    Returns:
        The data printed to standard out by the subprocess.

    """
    # Convert all command arguments to strings
    cmd = list(map(str, cmd))

    # Convert data to bytes
    if isinstance(data, str):
        data = data.encode()

    if use_tempfile:
        # If the command cannot read from stdin, create a temporary file to pass to
        # the command
        tfile = tempfile.NamedTemporaryFile(delete=False)
        tfile.write(data)
        tfile.close()
        cmd.append(tfile.name)

    # TODO render asynchronously
    # proc = await asyncio.create_subprocess_shell(
    # " ".join(cmd),
    # stdout=asyncio.subprocess.PIPE,
    # stdin=asyncio.subprocess.PIPE,
    # stderr=asyncio.subprocess.DEVNULL,
    # )
    # stdout, stderr = await proc.communicate(data)

    log.debug("Running external command `%s`", cmd)
    error: "Optional[Exception]" = None
    try:
        # Execution of untrusted input has been checked for
        output_bytes = subprocess.check_output(  # noqa S603
            cmd, input=data, stderr=subprocess.DEVNULL
        )
    except FileNotFoundError as error_:
        log.error("Could not run external command `%s`", cmd)
        error = error_
    except subprocess.CalledProcessError as error_:
        log.error("There was an error while running external command `%s`", cmd)
        error = error_
    finally:
        if error is not None:
            # Generate an output stating there was an error
            output_bytes = (
                b"\x1b[33m"  # Set fg to yellow
                b"\xee\x82\xb6"  # Draw left pill side
                b"\x1b[43m\x1b[30m"  # Set fg to black, bg to yellow
                b"\xe2\x9a\xa0"  # Draw warning symbol
                b" Rendering Error"
                b"\x1b[33m\x1b[49m"  # Set fg to yellow, reset bg
                b"\xee\x82\xb4"  # Draw right pill side
                b"\x1b[n"  # Reset style
            )

        # TODO Log any stderr

        # Clean up any temporary file
        if use_tempfile:
            tfile.close()
            os.unlink(tfile.name)

    return output_bytes


def data_pixel_size(
    data: "Any",
    format_: "str",
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
) -> "tuple[Optional[int], Optional[int]]":
    """Get the dimensions of an image.

    Foreground and background color are set at this point if they are available, as
    data conversion outputs are cached and re-used.

    Args:
        data: The data to check the dimensions of
        format_: The current format of the data
        fg: The desired foreground color of the data
        bg: The desired background color of the data

    Returns:
        A tuple of the data's width in terminal columns and its aspect ratio, when
            converted to a image.

    """
    px = py = None
    # Do not bother trying if the format is ANSI
    if format_ == "ansi":
        return px, py
    # Try using imagesize to get the size of the output
    if format_ not in {"png", "svg", "jpg", "gif", "tiff"}:
        try:
            data = convert(data, from_=format_, to="png", fg=fg, bg=bg)
        except NotImplementedError:
            pass
    if isinstance(data, str):
        data = data.encode()
    try:
        px_calc, py_calc = imagesize.get(io.BytesIO(data))
    except ValueError:
        pass
    else:
        if px_calc > 0:
            px = px_calc
        if py_calc > 0:
            py = py_calc
    return px, py


def pixels_to_cell_size(
    px: "Optional[int]" = None,
    py: "Optional[int]" = None,
) -> "tuple[int, float]":
    """Get the cell width and aspect ration of a pixel dimension.

    Args:
        px: The desired pixel width of the data if known
        py: The pixel height of the data if known

    Returns:
        A tuple of the data's width in terminal columns and its aspect ratio, when
            converted to a image.

    """
    cols, aspect = 0, 0.0
    if px is not None and py is not None:
        cell_px, cell_py = get_app().term_info.cell_size_px
        cols = max(1, int(px // cell_px))
        aspect = (py / cell_py) / (px / cell_px)
    return cols, aspect



"""Launch the euporie application."""


def main() -> "None":
    """Call the main entrypoint to the application."""
    from euporie.core import __main__

    __main__.main(__name__.split(".")[1])


if __name__ == "__main__":
    main()



"""Defines a simple app for launching euporie apps."""

from __future__ import annotations

from importlib.metadata import entry_points

from euporie.core.config import Config, add_setting

APP_ALIASES = {
    "edit": "notebook",
}


class CoreApp:
    """Launch a euporie application."""

    config = Config()

    @classmethod
    def launch(cls) -> "None":
        """Launches the app."""
        # Load the launcher's configuration
        cls.config.load(cls)
        app = cls.config.app

        # Add aliases
        app = APP_ALIASES.get(app, app)

        # Run the application
        from euporie.core.__main__ import main

        main(app)

    # ################################## Settings #####################################

    add_setting(
        name="app",
        flags=["app"],
        default="notebook",
        type_=str,
        required=False,
        help_="The application to launch",
        choices=sorted(
            {entry.name for entry in entry_points()["euporie.apps"]} - {"launch"}
            | APP_ALIASES.keys()
        ),
        description="""
            The name of the application to launch.
        """,
    )
